# -*- coding: utf-8 -*-

class Tokens:
    # objects

    Class = 'class'
    Enum = 'enum'

    # simple types

    Boolean = 'boolean'
    Char = 'char'
    Byte = 'byte'
    UnsignedByte = 'ubyte'
    Short = 'short'
    UnsignedShort = 'ushort'
    Integer = 'int'
    UnsignedInteger = 'uint'
    Long = 'long'
    UnsignedLong = 'ulong'
    Float = 'float'
    Double = 'double'
    String = 'string'

    # complex types

    List = 'list'
    Map = 'map'
    Array = 'array'



class Patterns:
    # objects

    Class = "%s(\((([a-zA-Z_][a-zA-Z0-9_]*,)*([a-zA-Z_][a-zA-Z0-9_]*))\))?$" % Tokens.Class
    Enum = "%s<(%s)>{(([a-zA-Z_][a-zA-Z0-9_]*(\(-?[0-9]+\))?,)*([a-zA-Z_][a-zA-Z0-9_]*(\(-?[0-9]+\))?))}$" % \
            (Tokens.Enum, '|'.join([
                Tokens.Byte, Tokens.UnsignedByte, Tokens.Short, Tokens.UnsignedShort,
                Tokens.Integer, Tokens.UnsignedInteger, Tokens.Long, Tokens.UnsignedLong
            ]))
    EnumValue = "([a-zA-Z_][a-zA-Z0-9_]*)(\((-?[0-9]+)\))?$"

    # simple types

    Boolean = "%s$" % Tokens.Boolean
    Char = "%s$" % Tokens.Char
    Byte = "%s$" % Tokens.Byte
    UnsignedByte = "%s$" % Tokens.UnsignedByte
    Short = "%s$" % Tokens.Short
    UnsignedShort = "%s$" % Tokens.UnsignedShort
    Integer = "%s$" % Tokens.Integer
    UnsignedInteger = "%s$" % Tokens.UnsignedInteger
    Long = "%s$" % Tokens.Long
    UnsignedLong = "%s$" % Tokens.UnsignedLong
    Float = "%s$" % Tokens.Float
    Double = "%s$" % Tokens.Double
    String = "%s$" % Tokens.String

    # complex types

    List = "%s<(.*)>$" % Tokens.List
    Map = "%s<(.*,.*)>$" % Tokens.Map
    Array = "%s((\[[0-9]+\])+)<(.*)>$" % Tokens.Array



class Rules:

    def __init__(self, regex_factory, get_type_tree):
        self._simples_token_list = [
            Tokens.Boolean,
            Tokens.Char,
            Tokens.Byte,
            Tokens.UnsignedByte,
            Tokens.Short,
            Tokens.UnsignedShort,
            Tokens.Integer,
            Tokens.UnsignedInteger,
            Tokens.Long,
            Tokens.UnsignedLong,
            Tokens.Float,
            Tokens.Double,
            Tokens.String
        ]

        self._simples_pattern_list = [
            'Boolean',
            'Char',
            'Byte',
            'UnsignedByte',
            'Short',
            'UnsignedShort',
            'Integer',
            'UnsignedInteger',
            'Long',
            'UnsignedLong',
            'Float',
            'Double',
            'String'
        ]

        self._regex_factory = regex_factory
        self._get_type_tree = get_type_tree


    # objects

    def _rule_class(self, text):
        result = self._regex_factory.match('Class', text)
        if result is None:
            return None

        parent_classes = result.group(2)
        if parent_classes is None:
            return (Tokens.Class, )

        parent_trees = []
        for parent in parent_classes.split(','):
            parent_trees.append(self._get_type_tree(parent))
        return (Tokens.Class, tuple(parent_trees))


    def _rule_enum(self, text):
        result = self._regex_factory.match('Enum', text)
        if result is None:
            return None

        value_type = self._get_type_tree(result.group(1))
        members = []

        for member_text in result.group(2).split(','):
            r = self._regex_factory.match('EnumValue', member_text)
            name = r.group(1)
            value = r.group(3)
            member = (name, ) if value is None else (name, int(value))
            members.append(member)
        return (Tokens.Enum, (value_type, tuple(members)))


    # simple types

    def _rule_simples(self, text):
        for i in range(len(self._simples_token_list)):
            result = self._regex_factory.match(self._simples_pattern_list[i], text)
            if result is None:
                continue
            return (self._simples_token_list[i], )
        return None


    # complex types

    def _rule_list(self, text):
        result = self._regex_factory.match('List', text)
        if result is None:
            return None

        inner_text = result.group(1)
        return (Tokens.List, self._get_type_tree(inner_text))


    def _rule_map(self, text):
        result = self._regex_factory.match('Map', text)
        if result is None:
            return None

        text = result.group(1)
        result = None

        diff = 0
        for i in range(len(text)):
            if text[i] == '<':
                diff += 1
            elif text[i] == '>':
                diff -= 1
            elif text[i] == ',' and diff == 0:
                result = (text[:i], text[i + 1:])
                break

        if result is None:
            return None

        inner_text_left = result[0]
        inner_text_right = result[1]
        return (Tokens.Map, (self._get_type_tree(inner_text_left), self._get_type_tree(inner_text_right)))


    def _rule_array(self, text):
        result = self._regex_factory.match('Array', text)
        if result is None:
            return None

        array_type = self._get_type_tree(result.group(3))
        dims = [int(r) for r in result.group(1)[1:-1].split('][')]

        if 0 in dims:
            return None
        return (Tokens.Array, (tuple(dims), array_type))



def convert_camel_case(name, is_lower):
    name = list(name)
    first_underscore_index = -1
    while name[first_underscore_index + 1] == '_':
        first_underscore_index += 1

    for i in range(len(name) - 2, first_underscore_index, -1):
        if name[i] == '_' and name[i + 1] not in ['_', '']:
            name[i + 1] = name[i + 1].upper()
            name[i] = ''
    for i in range(len(name)):
        if name[i] not in ['_', '']:
            if is_lower:
                name[i] = name[i].lower()
            else:
                name[i] = name[i].upper()
            break
    return ''.join(name)


capitalization_rules = {
    'snake_case': lambda name: name,
    'camelCase': lambda name: convert_camel_case(name, True),
    'PascalCase': lambda name: convert_camel_case(name, False)
}



# -*- coding: utf-8 -*-

# python imports
import sys
import struct
from enum import Enum

PY3 = sys.version_info > (3,)


class Parent(object):

	@staticmethod
	def name():
		return 'Parent'


	def __init__(self, count=None):
		self.initialize(count)
	

	def initialize(self, count=None):
		self.count = count
	

	def serialize(self):
		s = b''
		
		# serialize self.count
		s += b'\x00' if self.count is None else b'\x01'
		if self.count is not None:
			s += struct.pack('I', self.count)
		
		return s
	

	def deserialize(self, s, offset=0):
		# deserialize self.count
		tmp0 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp0:
			self.count = struct.unpack('I', s[offset:offset + 4])[0]
			offset += 4
		else:
			self.count = None
		
		return offset


class Child(Parent):

	@staticmethod
	def name():
		return 'Child'


	def __init__(self, count=None):
		self.initialize(count)
	

	def initialize(self, count=None):
		Parent.initialize(self, count)
	

	def serialize(self):
		s = b''
		
		# serialize parents
		s += Parent.serialize(self)
		
		return s
	

	def deserialize(self, s, offset=0):
		# deserialize parents
		offset = Parent.deserialize(self, s, offset)
		
		return offset



# -*- coding: utf-8 -*-


class CodeEditor(object):

    def __init__(self):
        self._code = ""
        self._global_indent = 0
        self._temp_variable_number = 0


    def get_code(self):
        while self._code[-1] == '\n':
            self._code = self._code[:-1]
        self._code += '\n'
        return self._code


    def add_line(self, text="", local_indent=0):
        local_indent += self._global_indent
        self._code += '\t' * local_indent + text + '\n'


    def add_lines(self, lines, local_indent=0):
        for text in lines:
            self.add_line(text, local_indent)


    def add_new_line(self, count=1):
        self._code += '\n' * count


    def increase_indentation(self, count=1):
        self._global_indent += count


    def decrease_indentation(self, count=1):
        self._global_indent -= count


    def new_tempvar(self):
        var = 'tmp%s' % self._temp_variable_number
        self._temp_variable_number += 1
        return var


    def clear(self):
        self._code = ""
        self._global_indent = 0
        self._temp_variable_number = 0



# -*- coding: utf-8 -*-

# python imports
import os
import errno
from copy import deepcopy

# project imports
from ..parser_core import Tokens, capitalization_rules
from .python import PythonCodeGenerator
from .cpp import CppCodeGenerator
from .cs import CsCodeGenerator
from .java import JavaCodeGenerator


class CodeGenerator:

    def __init__(self, source_path, programming_language, destination_dir, capitalization_rule):
        self._generators = {
            'python': PythonCodeGenerator(),
            'cpp': CppCodeGenerator(),
            'cs': CsCodeGenerator(),
            'java': JavaCodeGenerator()
        }

        self._module_name = os.path.splitext(os.path.basename(source_path))[0]
        self._programming_language = programming_language
        self._destination_dir = os.path.realpath(destination_dir)
        self._capitalization_rule = capitalization_rule


    def generate(self, parse_tree):
        parse_tree = self._apply_capitalization_rule(parse_tree)
        result = self._generators[self._programming_language].generate(parse_tree, self._module_name)
        for code, filename in result:
            path = os.path.join(self._destination_dir, filename)
            # os.makedirs(os.path.dirname(path), exist_ok=True)
            try:
                os.makedirs(os.path.dirname(path))
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
            with open(path, 'w') as f:
                f.write(code)


    def _apply_capitalization_rule(self, parse_tree):
        parse_tree = deepcopy(parse_tree)
        for user_defined_type, attribute_trees in parse_tree:
            if attribute_trees['_def'][0] == Tokens.Class:
                new_attr_names = []
                for attr_name in attribute_trees:
                    if attr_name != '_def':
                        new_attr_name = capitalization_rules[self._capitalization_rule](attr_name)
                        new_attr_names.append((attr_name, new_attr_name))
                for attr_name, new_attr_name in new_attr_names:
                    attribute_trees[new_attr_name] = attribute_trees.pop(attr_name)
        return parse_tree



# -*- coding: utf-8 -*-

# project imports
from .cpp_core import HeaderGenerator, FooterGenerator, TypeGenerator
from .code_editor import CodeEditor


class CppCodeGenerator:

    def __init__(self):
        self._code_editor = CppCodeEditor()
        self._header_generator = HeaderGenerator(self._code_editor)
        self._type_generator = TypeGenerator(self._code_editor)
        self._footer_generator = FooterGenerator(self._code_editor)


    def generate(self, parse_tree, module_name):
        self._generate(parse_tree, module_name)
        filename = '%s.h' % module_name
        return [(self._code_editor.get_code(), filename)]


    def _generate(self, parse_tree, module_name):
        self._header_generator.gen_header(module_name)
        for user_defined_type, properties in parse_tree:
            self._type_generator.gen_type(user_defined_type, properties)
            self._code_editor.add_line()
            if parse_tree.index((user_defined_type, properties)) < len(parse_tree) - 1:
                self._code_editor.add_line()
        self._footer_generator.gen_footer(module_name)



class CppCodeEditor(CodeEditor):

    def increase_indentation(self, put_accolade=True):
        if put_accolade:
            self.add_line('{')
        super(CppCodeEditor, self).increase_indentation()


    def decrease_indentation(self, put_accolade=True, add_semi=False):
        super(CppCodeEditor, self).decrease_indentation()
        if put_accolade:
            semi = ';' if add_semi else ''
            self.add_line('}' + semi)



# -*- coding: utf-8 -*-

# project imports
from ..parser_core import capitalization_rules
from .cs_core import RootGenerator, HeaderGenerator, FooterGenerator, TypeGenerator
from .cpp import CppCodeEditor


class CsCodeGenerator:

    def __init__(self):
        self._root_code_editor = CsCodeEditor()
        self._root_generator = RootGenerator(self._root_code_editor)

        self._code_editor = CsCodeEditor()
        self._header_generator = HeaderGenerator(self._code_editor)
        self._type_generator = TypeGenerator(self._code_editor)
        self._footer_generator = FooterGenerator(self._code_editor)


    def generate(self, parse_tree, module_name):
        module_name = capitalization_rules['PascalCase'](module_name)
        self._generate_root()
        self._generate(parse_tree, module_name)
        root_filename = 'KSObject.cs'
        filename = '%s.cs' % module_name
        return [
            (self._root_code_editor.get_code(), root_filename),
            (self._code_editor.get_code(), filename)
        ]


    def _generate(self, parse_tree, module_name):
        self._header_generator.gen_header(module_name)
        for user_defined_type, properties in parse_tree:
            self._type_generator.gen_type(user_defined_type, properties)
            if parse_tree.index((user_defined_type, properties)) < len(parse_tree) - 1:
                self._code_editor.add_line()
        self._footer_generator.gen_footer(module_name)


    def _generate_root(self):
        self._root_generator.gen_root()


class CsCodeEditor(CppCodeEditor):

    pass



# -*- coding: utf-8 -*-

# python imports
import struct

# project imports
from ..parser_core import Tokens


class HeaderGenerator:

    def __init__(self, code_editor):
        self._code_editor = code_editor


    def gen_header(self):
        self._code_editor.add_line("# -*- coding: utf-8 -*-\n")
        self._gen_imports()
        self._code_editor.add_line('\n')


    def _gen_imports(self):
        code_editor = self._code_editor

        code_editor.add_line("# python imports")
        code_editor.add_line("import sys")
        code_editor.add_line("import struct")
        code_editor.add_line("from enum import Enum")
        code_editor.add_line()
        code_editor.add_line("PY3 = sys.version_info > (3,)")


###################################################################################
###################################################################################


class TypeGenerator:

    def __init__(self, code_editor):
        self._user_defined_types = {
            Tokens.Class: {},
            Tokens.Enum: {}
        }

        self._code_editor = code_editor
        self._serializer_generator = SerializerGenerator(self._user_defined_types, code_editor)
        self._deserializer_generator = DeserializerGenerator(self._user_defined_types, code_editor)


    def gen_type(self, type_name, properties):
        gen_func = getattr(self, '_gen_type_%s' % properties['_def'][0])
        gen_func(type_name, properties)



    def _gen_type_class(self, type_name, properties):
        code_editor = self._code_editor

        _def = properties.pop('_def')
        properties = properties.items()

        parents = ['object'] if len(_def) == 1 else [p[0] for p in _def[1]]
        self._user_defined_types[Tokens.Class][type_name] = (parents, properties)

        # generate definitions
        code_editor.add_line("class %s(%s):" % (type_name, ', '.join(parents)))
        code_editor.add_line()

        # generate name
        code_editor.add_line("@staticmethod", 1)
        code_editor.add_line("def name():", 1)
        code_editor.add_line("return '%s'" % type_name, 2)

        code_editor.add_line('\n')
        code_editor.increase_indentation()

        # generate constructor
        attrs = self._get_attrs(type_name)
        parameters = ', '.join(['self'] + ["%s=None" % a for a in attrs])
        code_editor.add_line("def __init__(%s):" % parameters)
        code_editor.add_line("self.initialize(%s)" % ', '.join(attrs), 1)
        code_editor.add_line('\n')

        # generate initializer
        code_editor.add_line("def initialize(%s):" % parameters)
        code_editor.increase_indentation()

        is_empty_initialize = True
        if parents[0] != 'object':
            is_empty_initialize = False
            for parent in parents:
                parent_attrs = self._get_attrs(parent)
                code_editor.add_line("%s.initialize(%s)" % (parent, ', '.join(['self'] + parent_attrs)))

        if properties:
            if not is_empty_initialize:
                code_editor.add_line()
            is_empty_initialize = False
            for attr_name, tree in properties:
                self._code_editor.add_line("self.%s = %s" % (attr_name, attr_name))

        if is_empty_initialize:
            code_editor.add_line('pass')

        code_editor.decrease_indentation()
        code_editor.add_line('\n')

        # generate serializer
        result_name = 's'
        code_editor.add_line("def serialize(self):")
        code_editor.increase_indentation()
        code_editor.add_line("%s = b''" % result_name)
        code_editor.add_line()

        if parents[0] != 'object':
            code_editor.add_line("# serialize parents")
            for parent in parents:
                code_editor.add_line("%s += %s.serialize(self)" % (result_name, parent))
            code_editor.add_line()

        for attr_name, tree in properties:
            attr_name = "self.%s" % attr_name
            code_editor.add_line("# serialize %s" % attr_name)
            self._serializer_generator.gen_serializer(attr_name, result_name, tree)
            code_editor.add_line()

        code_editor.add_line("return %s" % result_name)
        code_editor.decrease_indentation()
        code_editor.add_line('\n')

        # generate deserializer
        data_name = 's'
        offset_name = 'offset'
        code_editor.add_line("def deserialize(self, %s, %s=0):" % (data_name, offset_name))
        code_editor.increase_indentation()

        if parents[0] != 'object':
            code_editor.add_line("# deserialize parents")
            for parent in parents:
                code_editor.add_line("%s = %s.deserialize(self, %s, %s)" % 
                                     (offset_name, parent, data_name, offset_name))
            code_editor.add_line()

        for attr_name, tree in properties:
            attr_name = "self.%s" % attr_name
            code_editor.add_line("# deserialize %s" % attr_name)
            self._deserializer_generator.gen_deserializer(data_name, offset_name, attr_name, tree)
            code_editor.add_line()

        code_editor.add_line("return %s" % offset_name)
        code_editor.decrease_indentation()

        code_editor.decrease_indentation()


    def _gen_type_enum(self, type_name, properties):
        code_editor = self._code_editor

        _def = properties.pop('_def')
        value_type = _def[1][0]
        self._user_defined_types[Tokens.Enum][type_name] = value_type
        code_editor.add_line("class %s(Enum):" % type_name)

        value = 0
        for attr in _def[1][1]:
            name = attr[0]
            if len(attr) > 1:
                value = attr[1]
            code_editor.add_line("%s = %s" % (name, value), 1)
            value += 1


    def _get_attrs(self, type_name):
        if type_name == 'object':
            return []
        parents, properties = self._user_defined_types[Tokens.Class][type_name]
        result = []
        for p in parents:
            result += self._get_attrs(p)
        result += [p for p, _ in properties]
        return result


###################################################################################
###################################################################################


class SerializerGenerator:

    def __init__(self, user_defined_types, code_editor):
        self._simples_format_table = {
            Tokens.Boolean: '?',
            Tokens.Char: 'c',
            Tokens.Byte: 'b',
            Tokens.UnsignedByte: 'B',
            Tokens.Short: 'h',
            Tokens.UnsignedShort: 'H',
            Tokens.Integer: 'i',
            Tokens.UnsignedInteger: 'I',
            Tokens.Long: 'q',
            Tokens.UnsignedLong: 'Q',
            Tokens.Float: 'f',
            Tokens.Double: 'd'
        }

        self._user_defined_types = user_defined_types
        self._code_editor = code_editor


    def gen_serializer(self, value_name, result_name, tree):
        value_type = tree[0]
        serializer_func = None

        for type_token in self._user_defined_types:
            if value_type in self._user_defined_types[type_token]:
                serializer_func = getattr(self, '_gen_serializer_%s' % type_token)
                break

        if serializer_func is None:
            serializer_func = getattr(
                self,
                '_gen_serializer_%s' % value_type,
                self._gen_serializer_simples
            )

        self._code_editor.add_line("%s += b'\\x00' if %s is None else b'\\x01'" % 
                                   (result_name, value_name))
        self._code_editor.add_line("if %s is not None:" % value_name)
        self._code_editor.increase_indentation()
        serializer_func(value_name, result_name, tree)
        self._code_editor.decrease_indentation()



    def _gen_serializer_class(self, value_name, result_name, tree):
        self._code_editor.add_line("%s += %s.serialize()" % (result_name, value_name))


    def _gen_serializer_enum(self, value_name, result_name, tree):
        self._gen_serializer_simples(
            "%s.value" % value_name,
            result_name,
            self._user_defined_types[Tokens.Enum][tree[0]]
        )


    def _gen_serializer_list(self, value_name, result_name, tree):
        code_editor = self._code_editor

        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen items
        item = code_editor.new_tempvar()
        code_editor.add_line("for %s in %s:" % (item, value_name))
        code_editor.increase_indentation()
        self.gen_serializer(item, result_name, tree[1])
        code_editor.decrease_indentation()


    def _gen_serializer_map(self, value_name, result_name, tree):
        code_editor = self._code_editor

        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen items
        key = code_editor.new_tempvar()
        code_editor.add_line("for %s in %s:" % (key, value_name))
        code_editor.increase_indentation()
        self.gen_serializer(key, result_name, tree[1][0])
        self.gen_serializer("%s[%s]" % (value_name, key), result_name, tree[1][1])
        code_editor.decrease_indentation()


    def _gen_serializer_array(self, value_name, result_name, tree):
        code_editor = self._code_editor

        dims = tree[1][0]
        array_type = tree[1][1]

        indexes = []
        for i in range(len(dims)):
            index = code_editor.new_tempvar()
            indexes.append(index)
            code_editor.add_line("for %s in range(%s):" % (index, dims[i]), i)

        code_editor.increase_indentation(len(dims))
        value_name += "[%s]" % ']['.join(indexes)
        self.gen_serializer(value_name, result_name, array_type)
        code_editor.decrease_indentation(len(dims))


    def _gen_serializer_string(self, value_name, result_name, tree):
        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen string
        self._code_editor.add_line("%s += %s.encode('ISO-8859-1') if PY3 else %s" % 
                                   (result_name, value_name, value_name))


    def _gen_serializer_char(self, value_name, result_name, tree):
        value_name = "%s.encode('ISO-8859-1') if PY3 else %s" % (value_name, value_name)
        self._gen_serializer_simples(value_name, result_name, tree)


    def _gen_serializer_simples(self, value_name, result_name, tree):
        fmt = self._simples_format_table[tree[0]]
        self._code_editor.add_line("%s += struct.pack('%s', %s)" % (result_name, fmt, value_name))


    def _gen_size(self, value_name, result_name, tree):
        code_editor = self._code_editor

        size = code_editor.new_tempvar()
        code_editor.add_line("%s = b''" % size)
        self._gen_serializer_simples("len(%s)" % value_name, size, (Tokens.UnsignedInteger, ))
        code_editor.add_line("while len(%s) and %s[-1] == b'\\x00'[0]:" % (size, size))
        code_editor.add_line("%s = %s[:-1]" % (size, size), 1)

        self._gen_serializer_simples("len(%s)" % size, result_name, (Tokens.UnsignedByte, ))
        code_editor.add_line("%s += %s" % (result_name, size))
        code_editor.add_line()


###################################################################################
###################################################################################


class DeserializerGenerator:

    def __init__(self, user_defined_types, code_editor):
        self._simples_format_table = {
            Tokens.Boolean: '?',
            Tokens.Char: 'c',
            Tokens.Byte: 'b',
            Tokens.UnsignedByte: 'B',
            Tokens.Short: 'h',
            Tokens.UnsignedShort: 'H',
            Tokens.Integer: 'i',
            Tokens.UnsignedInteger: 'I',
            Tokens.Long: 'q',
            Tokens.UnsignedLong: 'Q',
            Tokens.Float: 'f',
            Tokens.Double: 'd'
        }

        self._simples_fmtsize_table = {
            key: struct.calcsize(val) for key, val in self._simples_format_table.items()
        }

        self._user_defined_types = user_defined_types
        self._code_editor = code_editor



    def _get_data(self, data_name, offset_name, length):
        return "%s[%s:%s + %s]" % (data_name, offset_name, offset_name, length)

    def _increase_offset(self, offset_name, n):
        self._code_editor.add_line("%s += %s" % (offset_name, n))



    def gen_deserializer(self, data_name, offset_name, value_name, tree):
        value_type = tree[0]
        deserializer_func = None

        for type_token in self._user_defined_types:
            if value_type in self._user_defined_types[type_token]:
                deserializer_func = getattr(self, '_gen_deserializer_%s' % type_token)
                break

        if deserializer_func is None:
            deserializer_func = getattr(
                self,
                '_gen_deserializer_%s' % value_type,
                self._gen_deserializer_simples
            )

        is_null = self._code_editor.new_tempvar()
        self._gen_deserializer_simples(data_name, offset_name, is_null, (Tokens.UnsignedByte, ))
        self._code_editor.add_line("if %s:" % is_null)
        self._code_editor.increase_indentation()
        deserializer_func(data_name, offset_name, value_name, tree)
        self._code_editor.decrease_indentation()
        self._code_editor.add_line("else:")
        self._code_editor.add_line("%s = None" % value_name, 1)



    def _gen_deserializer_class(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        code_editor.add_line("%s = %s()" % (value_name, tree[0]))
        code_editor.add_line("%s = %s.deserialize(%s, %s)" % 
                             (offset_name, value_name, data_name, offset_name))


    def _gen_deserializer_enum(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        enum_type = tree[0]
        val = code_editor.new_tempvar()
        self._gen_deserializer_simples(
            data_name,
            offset_name,
            val,
            self._user_defined_types[Tokens.Enum][tree[0]]
        )
        code_editor.add_line("%s = %s(%s)" % (value_name, enum_type, val))


    def _gen_deserializer_list(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen items
        code_editor.add_line("%s = []" % value_name)
        i = code_editor.new_tempvar()
        code_editor.add_line("for %s in range(%s):" % (i, size))
        item = code_editor.new_tempvar()
        code_editor.increase_indentation()
        self.gen_deserializer(data_name, offset_name, item, tree[1])
        code_editor.add_line("%s.append(%s)" % (value_name, item))
        code_editor.decrease_indentation()


    def _gen_deserializer_map(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen items
        code_editor.add_line("%s = {}" % value_name)
        i = code_editor.new_tempvar()
        code_editor.add_line("for %s in range(%s):" % (i, size))
        key = code_editor.new_tempvar()
        val = code_editor.new_tempvar()
        code_editor.increase_indentation()
        self.gen_deserializer(data_name, offset_name, key, tree[1][0])
        self.gen_deserializer(data_name, offset_name, val, tree[1][1])
        code_editor.add_line("%s[%s] = %s" % (value_name, key, val))
        code_editor.decrease_indentation()


    def _gen_deserializer_array(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        dims = tree[1][0]
        array_type = tree[1][1]

        initial_value = 'None'
        for dim in dims:
            initial_value = "[%s for _ in range(%s)]" % (initial_value, dim)
        code_editor.add_line("%s = %s" % (value_name, initial_value))

        indexes = []
        for i in range(len(dims)):
            index = code_editor.new_tempvar()
            indexes.append(index)
            code_editor.add_line("for %s in range(%s):" % (index, dims[i]), i)

        code_editor.increase_indentation(len(dims))
        value_name += "[%s]" % ']['.join(indexes)
        self.gen_deserializer(data_name, offset_name, value_name, array_type)
        code_editor.decrease_indentation(len(dims))


    def _gen_deserializer_string(self, data_name, offset_name, value_name, tree):
        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen string
        data = self._get_data(data_name, offset_name, size)
        self._code_editor.add_line("%s = %s.decode('ISO-8859-1') if PY3 else %s" % 
                                   (value_name, data, data))
        self._increase_offset(offset_name, size)


    def _gen_deserializer_char(self, data_name, offset_name, value_name, tree):
        self._gen_deserializer_simples(data_name, offset_name, value_name, tree)
        self._code_editor.add_line("%s = %s.decode('ISO-8859-1') if PY3 else %s" %
                                   (value_name, value_name, value_name))


    def _gen_deserializer_simples(self, data_name, offset_name, value_name, tree):
        fmt = self._simples_format_table[tree[0]]
        size = self._simples_fmtsize_table[tree[0]]
        data = data_name if offset_name is None else self._get_data(data_name, offset_name, size)
        self._code_editor.add_line("%s = struct.unpack('%s', %s)[0]" % 
                                   (value_name, fmt, data))
        if offset_name is not None:
            self._increase_offset(offset_name, size)


    def _gen_size(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        size_len = code_editor.new_tempvar()
        self._gen_deserializer_simples(data_name, offset_name, size_len, (Tokens.UnsignedByte, ))
        size_bytes = code_editor.new_tempvar()
        code_editor.add_line("%s = %s" % (size_bytes, self._get_data(data_name, offset_name, size_len)))
        self._increase_offset(offset_name, size_len)
        code_editor.add_line("%s += b'\\x00' * (%s - %s)" % 
                             (size_bytes, self._simples_fmtsize_table[Tokens.UnsignedInteger], size_len))
        size = code_editor.new_tempvar()
        self._gen_deserializer_simples(size_bytes, None, size, (Tokens.UnsignedInteger, ))
        code_editor.add_line()

        return size



# -*- coding: utf-8 -*-

# project imports
from .python_core import HeaderGenerator, TypeGenerator
from .code_editor import CodeEditor


class PythonCodeGenerator:

    def __init__(self):
        self._code_editor = PythonCodeEditor()
        self._header_generator = HeaderGenerator(self._code_editor)
        self._type_generator = TypeGenerator(self._code_editor)


    def generate(self, parse_tree, module_name):
        self._generate(parse_tree)
        filename = '%s.py' % module_name
        return [(self._code_editor.get_code(), filename)]


    def _generate(self, parse_tree):
        self._header_generator.gen_header()
        for user_defined_type, properties in parse_tree:
            self._type_generator.gen_type(user_defined_type, properties)
            self._code_editor.add_line('\n')



class PythonCodeEditor(CodeEditor):

    pass



# -*- coding: utf-8 -*-

# project imports
from .parser import Parser
from .code_generators import CodeGenerator


def generate(source_path, programming_language, destination_dir, capitalization_rule):
    parser = Parser(source_path)
    code_generator = CodeGenerator(source_path, programming_language, destination_dir, capitalization_rule)
    code_generator.generate(parser.parse_tree())



# -*- coding: utf-8 -*-

# python imports
import warnings

# project imports
from ..parser_core import Tokens, capitalization_rules


class RootGenerator:

    def __init__(self, code_editor):
        self._code_editor = code_editor


    def gen_root(self):
        code_editor = self._code_editor

        self._gen_package('ks')
        code_editor.add_line()
        self._gen_imports()
        code_editor.add_line()
        self._gen_ks_object()


    def _gen_imports(self):
        self._code_editor.add_line("import java.lang.*;")
        self._code_editor.add_line("import java.util.*;")


    def _gen_ks_object(self):
        self._code_editor.add_lines([
            "public abstract class KSObject",
            "{",
            "\tpublic static final String nameStatic = \"\";",
            "\tpublic abstract String name();",
            "\tpublic abstract byte[] serialize();",
            "\tpublic int deserialize(byte[] s) { return deserialize(s, 0); }",
            "\tprotected abstract int deserialize(byte[] s, int offset);",
            "",
            "",
			"\tprotected static List<Byte> b2B(byte[] bytes)",
			"\t{",
			"\t\tList<Byte> result = new ArrayList<>();",
			"\t\tfor (byte b : bytes)",
			"\t\t\tresult.add(b);",
			"\t\treturn result;",
			"\t}",
            "",
			"\tprotected static byte[] B2b(List<Byte> bytes)",
			"\t{",
			"\t\tbyte[] result = new byte[bytes.size()];",
			"\t\tfor(int i = 0; i < result.length; i++)",
			"\t\t\tresult[i] = bytes.get(i);",
			"\t\treturn result;",
			"\t}",
            "}",
        ])


    def _gen_package(self, pkg_name):
        self._code_editor.add_line("package %s;" % pkg_name)


###################################################################################
###################################################################################


class HeaderGenerator:

    def __init__(self, code_editor):
        self._code_editor = code_editor


    def gen_header(self, module_name):
        code_editor = self._code_editor

        self._gen_package("ks." + module_name)
        code_editor.add_line()
        self._gen_imports()
        code_editor.add_line()


    def _gen_imports(self):
        self._code_editor.add_line("import java.lang.*;")
        self._code_editor.add_line("import java.util.*;")
        self._code_editor.add_line("import java.nio.*;")
        self._code_editor.add_line("import java.nio.charset.Charset;")
        self._code_editor.add_line()
        self._code_editor.add_line("import ks.KSObject;")


    def _gen_package(self, pkg_name):
        self._code_editor.add_line("package %s;" % pkg_name)


###################################################################################
###################################################################################


class FooterGenerator:

    def __init__(self, code_editor):
        self._code_editor = code_editor


    def gen_footer(self, module_name):
        self._code_editor.decrease_indentation(put_accolade=False)


###################################################################################
###################################################################################


class TypeGenerator:

    def __init__(self, code_editor):
        self._user_defined_types = {
            Tokens.Class: {},
            Tokens.Enum: {}
        }

        self._code_editor = code_editor
        self._serializer_generator = SerializerGenerator(self._user_defined_types, code_editor)
        self._deserializer_generator = DeserializerGenerator(self._user_defined_types, code_editor)
        self._variable_type_generator = VariableTypeGenerator(self._user_defined_types)


    def gen_type(self, type_name, properties):
        gen_func = getattr(self, '_gen_type_%s' % properties['_def'][0])
        gen_func(type_name, properties)



    def _gen_type_class(self, type_name, properties):
        code_editor = self._code_editor

        _def = properties.pop('_def')
        properties = properties.items()

        parents = ['KSObject'] if len(_def) == 1 else [p[0] for p in _def[1]]
        if len(parents) > 1:
            warnings.warn("Java does not support multiple inheritance. Only first parent will be considered.")
            parents = parents[:1]
        self._user_defined_types[Tokens.Class][type_name] = (parents, properties)

        # generate definitions
        code_editor.add_line("public class %s extends %s" % (type_name, ', '.join(parents)))
        code_editor.increase_indentation()
        for attr_name, tree in properties:
            vartype = self._variable_type_generator.gen_vartype(tree)
            if tree[0] in [Tokens.UnsignedByte, Tokens.UnsignedShort, Tokens.UnsignedInteger, Tokens.UnsignedLong]:
                warnings.warn("Java does not support unsigned types. Instead, signed types will be considered.")
            code_editor.add_line("protected %s %s;" % (vartype, attr_name))
        code_editor.add_line()

        # generate getters
        code_editor.add_line("// getters")
        code_editor.add_line()
        for attr_name, tree in properties:
            vartype = self._variable_type_generator.gen_vartype(tree)
            code_editor.add_line("public %s get%s()" % (vartype, capitalization_rules['PascalCase'](attr_name)))
            code_editor.increase_indentation()
            code_editor.add_line("return this.%s;" % attr_name)
            code_editor.decrease_indentation()
            code_editor.add_line()
        code_editor.add_line()

        # generate setters
        code_editor.add_line("// setters")
        code_editor.add_line()
        for attr_name, tree in properties:
            vartype = self._variable_type_generator.gen_vartype(tree)
            code_editor.add_line("public void set%s(%s %s)" % (capitalization_rules['PascalCase'](attr_name), vartype, attr_name))
            code_editor.increase_indentation()
            code_editor.add_line("this.%s = %s;" % (attr_name, attr_name))
            code_editor.decrease_indentation()
            code_editor.add_line()
        code_editor.add_line()

        # generate constructor
        code_editor.add_line("public %s()" % type_name)
        code_editor.increase_indentation()
        code_editor.decrease_indentation()
        code_editor.add_line()

        # generate name
        code_editor.add_line("public static final String nameStatic = \"%s\";" % type_name)
        code_editor.add_line()

        code_editor.add_line("@Override")
        code_editor.add_line("public String name() { return \"%s\"; }" % type_name)
        code_editor.add_line()

        # generate serializer
        result_name = 's'
        code_editor.add_line("@Override")
        code_editor.add_line("public byte[] serialize()")
        code_editor.increase_indentation()
        code_editor.add_line("List<Byte> %s = new ArrayList<>();" % result_name)
        code_editor.add_line()

        if parents[0] != 'KSObject':
            code_editor.add_line("// serialize parents")
            # for parent in parents:
            #     code_editor.add_line("%s.addAll(b2B(%s.serialize()));" % (result_name, parent))
            code_editor.add_line("%s.addAll(b2B(super.serialize()));" % result_name)
            code_editor.add_line()

        for attr_name, tree in properties:
            code_editor.add_line("// serialize %s" % attr_name)
            self._serializer_generator.gen_serializer(attr_name, result_name, tree)
            code_editor.add_line()

        code_editor.add_line("return B2b(%s);" % result_name)
        code_editor.decrease_indentation()
        code_editor.add_line()

        # generate deserializer
        data_name = 's'
        offset_name = 'offset'
        code_editor.add_line("@Override")
        code_editor.add_line("protected int deserialize(byte[] %s, int %s)" %
                             (data_name, offset_name))
        code_editor.increase_indentation()

        if parents[0] != 'KSObject':
            code_editor.add_line("// deserialize parents")
            # for parent in parents:
            #     code_editor.add_line("%s = %s.deserialize(%s, %s);" % 
            #                          (offset_name, parent, data_name, offset_name))
            code_editor.add_line("%s = super.deserialize(%s, %s);" % 
                                 (offset_name, data_name, offset_name))
            code_editor.add_line()

        for attr_name, tree in properties:
            code_editor.add_line("// deserialize %s" % attr_name)
            self._deserializer_generator.gen_deserializer(data_name, offset_name, attr_name, tree)
            code_editor.add_line()

        code_editor.add_line("return %s;" % offset_name)
        code_editor.decrease_indentation()

        code_editor.decrease_indentation()


    def _gen_type_enum(self, type_name, properties):
        code_editor = self._code_editor

        _def = properties.pop('_def')
        value_type = _def[1][0]
        self._user_defined_types[Tokens.Enum][type_name] = value_type
        code_editor.add_line("public enum %s" % type_name)
        code_editor.increase_indentation()

        value_type_cls = self._variable_type_generator.gen_vartype(value_type)
        value_type = self._variable_type_generator.gen_vartype(value_type, is_primitive=True)

        value = 0
        for attr in _def[1][1]:
            name = attr[0]
            if len(attr) > 1:
                value = attr[1]
            code_editor.add_line("%s((%s) %s)," % (name, value_type, value))
            value += 1
        code_editor.add_line(";\n")

        code_editor.add_lines([
            "private final %s value;" % value_type,
            "%s(%s value) { this.value = value; }" % (type_name, value_type),
            "public %s getValue() { return value; }" % value_type,
            "",
            "private static Map<%s, %s> reverseLookup;" % (value_type_cls, type_name),
            "",
            "public static %s of(%s value)" % (type_name, value_type),
            "{",
			"\tif (reverseLookup == null)",
			"\t{",
			"\t\treverseLookup = new HashMap<>();",
			"\t\tfor (%s c : %s.values())" % (type_name, type_name),
			"\t\t\treverseLookup.put(c.getValue(), c);",
			"\t}",
			"\treturn reverseLookup.get(value);",
            "}",
        ])
        code_editor.decrease_indentation()

###################################################################################
###################################################################################


class SerializerGenerator:

    def __init__(self, user_defined_types, code_editor):
        self._simples_token_table = {
            Tokens.Short: 'Short',
            Tokens.UnsignedShort: 'Short',
            Tokens.Integer: 'Integer',
            Tokens.UnsignedInteger: 'Integer',
            Tokens.Long: 'Long',
            Tokens.UnsignedLong: 'Long',
            Tokens.Float: 'Float',
            Tokens.Double: 'Double' 
        }
        self._simples_conversion_method_table = {
            Tokens.Short: 'Short',
            Tokens.UnsignedShort: 'Short',
            Tokens.Integer: 'Int',
            Tokens.UnsignedInteger: 'Int',
            Tokens.Long: 'Long',
            Tokens.UnsignedLong: 'Long',
            Tokens.Float: 'Float',
            Tokens.Double: 'Double' 
        }

        self._user_defined_types = user_defined_types
        self._code_editor = code_editor
        self._variable_type_generator = VariableTypeGenerator(self._user_defined_types)


    def gen_serializer(self, value_name, result_name, tree):
        value_type = tree[0]
        serializer_func = None

        for type_token in self._user_defined_types:
            if value_type in self._user_defined_types[type_token]:
                serializer_func = getattr(self, '_gen_serializer_%s' % type_token)
                break

        if serializer_func is None:
            serializer_func = getattr(
                self,
                '_gen_serializer_%s' % value_type,
                self._gen_serializer_simples
            )

        self._code_editor.add_line("%s.add((byte) ((%s == null) ? 0 : 1));" % 
                                   (result_name, value_name))
        self._code_editor.add_line("if (%s != null)" % value_name)
        self._code_editor.increase_indentation()
        serializer_func(value_name, result_name, tree)
        self._code_editor.decrease_indentation()



    def _gen_serializer_class(self, value_name, result_name, tree):
        self._code_editor.add_line("%s.addAll(b2B(%s.serialize()));" % (result_name, value_name))


    def _gen_serializer_enum(self, value_name, result_name, tree):
        self._gen_serializer_simples(
            "(%s.getValue())" % value_name,
            result_name,
            self._user_defined_types[Tokens.Enum][tree[0]]
        )


    def _gen_serializer_list(self, value_name, result_name, tree):
        code_editor = self._code_editor

        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen items
        value_type = self._variable_type_generator.gen_vartype(tree[1])
        item = code_editor.new_tempvar()
        code_editor.add_line("for (%s %s : %s)" % (value_type, item, value_name))
        code_editor.increase_indentation()
        self.gen_serializer(item, result_name, tree[1])
        code_editor.decrease_indentation()


    def _gen_serializer_map(self, value_name, result_name, tree):
        code_editor = self._code_editor

        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen items
        value_type = self._variable_type_generator.gen_vartype(tree).replace('Map', 'Map.Entry', 1)
        pair = code_editor.new_tempvar()
        code_editor.add_line("for (%s %s : %s.entrySet())" % (value_type, pair, value_name))
        code_editor.increase_indentation()
        self.gen_serializer("%s.getKey()" % pair, result_name, tree[1][0])
        code_editor.add_line()
        self.gen_serializer("%s.getValue()" % pair, result_name, tree[1][1])
        code_editor.decrease_indentation()


    def _gen_serializer_array(self, value_name, result_name, tree):
        code_editor = self._code_editor

        dims = tree[1][0]
        array_type = tree[1][1]

        indexes = []
        for dim in dims:
            index = code_editor.new_tempvar()
            indexes.append(index)
            code_editor.add_line("for (int %s = 0; %s < %s; %s++)" %
                                 (index, index, dim, index))
            code_editor.increase_indentation()

        value_name += "[%s]" % ']['.join(indexes)
        self.gen_serializer(value_name, result_name, array_type)
        for _ in range(len(dims)):
            code_editor.decrease_indentation()


    def _gen_serializer_string(self, value_name, result_name, tree):
        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen string
        self._code_editor.add_line("%s.addAll(b2B(%s.getBytes(Charset.forName(\"ISO-8859-1\"))));" %
                                    (result_name, value_name))


    def _gen_serializer_simples(self, value_name, result_name, tree):
        if tree[0] == Tokens.UnsignedByte:
            self._code_editor.add_line("%s.add((byte) %s);" % (result_name, value_name))
        elif tree[0] == Tokens.Byte:
            self._code_editor.add_line("%s.add((byte) %s);" % (result_name, value_name))
        elif tree[0] == Tokens.Boolean:
            self._code_editor.add_line("%s.add((byte) ((%s) ? 1 : 0));" % (result_name, value_name))
        elif tree[0] == Tokens.Char:
            self._code_editor.add_line("%s.add((byte) (char) %s);" % (result_name, value_name))
        else:
            value_type = self._simples_token_table[tree[0]]
            conversion_method = self._simples_conversion_method_table[tree[0]]
            self._code_editor.add_line("%s.addAll(b2B(ByteBuffer.allocate(%s.BYTES).order(ByteOrder.LITTLE_ENDIAN).put%s(%s).array()));" % 
                                        (result_name, value_type, conversion_method, value_name))


    def _gen_size(self, value_name, result_name, tree):
        code_editor = self._code_editor

        size_method = 'length' if tree[0] == Tokens.String else 'size'
        size = code_editor.new_tempvar()
        code_editor.add_line("List<Byte> %s = new ArrayList<>();" % size)
        self._gen_serializer_simples("%s.%s()" % (value_name, size_method), size, (Tokens.UnsignedInteger, ))
        code_editor.add_line("while (%s.size() > 0 && %s.get(%s.size() - 1) == 0)" % (size, size, size))
        code_editor.add_line("%s.remove(%s.size() - 1);" % (size, size), 1)

        self._gen_serializer_simples("%s.size()" % size, result_name, (Tokens.UnsignedByte, ))
        code_editor.add_line("%s.addAll(%s);" % (result_name, size))
        code_editor.add_line()


###################################################################################
###################################################################################


class DeserializerGenerator:

    def __init__(self, user_defined_types, code_editor):
        self._simples_token_table = {
            Tokens.Short: 'Short',
            Tokens.UnsignedShort: 'Short',
            Tokens.Integer: 'Integer',
            Tokens.UnsignedInteger: 'Integer',
            Tokens.Long: 'Long',
            Tokens.UnsignedLong: 'Long',
            Tokens.Float: 'Float',
            Tokens.Double: 'Double' 
        }
        self._simples_conversion_method_table = {
            Tokens.Short: 'Short',
            Tokens.UnsignedShort: 'Short',
            Tokens.Integer: 'Int',
            Tokens.UnsignedInteger: 'Int',
            Tokens.Long: 'Long',
            Tokens.UnsignedLong: 'Long',
            Tokens.Float: 'Float',
            Tokens.Double: 'Double' 
        }

        self._user_defined_types = user_defined_types
        self._code_editor = code_editor
        self._variable_type_generator = VariableTypeGenerator(self._user_defined_types)


    def gen_deserializer(self, data_name, offset_name, value_name, tree):
        value_type = tree[0]
        deserializer_func = None

        for type_token in self._user_defined_types:
            if value_type in self._user_defined_types[type_token]:
                deserializer_func = getattr(self, '_gen_deserializer_%s' % type_token)
                break

        if deserializer_func is None:
            deserializer_func = getattr(
                self,
                '_gen_deserializer_%s' % value_type,
                self._gen_deserializer_simples
            )

        
        is_null = self._code_editor.new_tempvar()
        self._code_editor.add_line("byte %s;" % is_null)
        self._gen_deserializer_simples(data_name, offset_name, is_null, (Tokens.UnsignedByte, ))
        self._code_editor.add_line("if (%s == 1)" % is_null)
        self._code_editor.increase_indentation()
        deserializer_func(data_name, offset_name, value_name, tree)
        self._code_editor.decrease_indentation()
        self._code_editor.add_line("else")
        self._code_editor.add_line("%s = null;" % value_name, 1)



    def _gen_deserializer_class(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        code_editor.add_line("%s = new %s();" % (value_name, tree[0]))
        code_editor.add_line("%s = %s.deserialize(%s, %s);" % 
                             (offset_name, value_name, data_name, offset_name))


    def _gen_deserializer_enum(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        enum_type = tree[0]
        val = code_editor.new_tempvar()
        val_type = self._user_defined_types[Tokens.Enum][tree[0]][0]
        val_type = self._variable_type_generator.gen_vartype((val_type, ), is_primitive=True)
        code_editor.add_line("%s %s;" % (val_type, val))

        self._gen_deserializer_simples(
            data_name,
            offset_name,
            val,
            self._user_defined_types[Tokens.Enum][tree[0]]
        )
        code_editor.add_line("%s = %s.of(%s);" % (value_name, enum_type, val))


    def _gen_deserializer_list(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen items
        code_editor.add_line("%s = new ArrayList<>();" % value_name)
        i = code_editor.new_tempvar()
        code_editor.add_line("for (int %s = 0; %s < %s; %s++)" % (i, i, size, i))
        code_editor.increase_indentation()
        item = code_editor.new_tempvar()
        item_type = self._variable_type_generator.gen_vartype(tree[1])
        code_editor.add_line("%s %s;" % (item_type, item))
        self.gen_deserializer(data_name, offset_name, item, tree[1])
        code_editor.add_line("%s.add(%s);" % (value_name, item))
        code_editor.decrease_indentation()


    def _gen_deserializer_map(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen items
        code_editor.add_line("%s = new HashMap<>();" % value_name)
        i = code_editor.new_tempvar()
        code_editor.add_line("for (int %s = 0; %s < %s; %s++)" % (i, i, size, i))
        code_editor.increase_indentation()

        key = code_editor.new_tempvar()
        val = code_editor.new_tempvar()
        key_type = self._variable_type_generator.gen_vartype(tree[1][0])
        val_type = self._variable_type_generator.gen_vartype(tree[1][1])

        code_editor.add_line("%s %s;" % (key_type, key))
        self.gen_deserializer(data_name, offset_name, key, tree[1][0])
        code_editor.add_line()

        code_editor.add_line("%s %s;" % (val_type, val))
        self.gen_deserializer(data_name, offset_name, val, tree[1][1])
        code_editor.add_line()

        code_editor.add_line("%s.put(%s, %s);" % (value_name, key, val))
        code_editor.decrease_indentation()


    def _gen_deserializer_array(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        dims = tree[1][0]
        array_type_tree = tree[1][1]
        array_type = self._variable_type_generator.gen_vartype(array_type_tree)
        if array_type.startswith('List'):
            array_type = 'ArrayList'
        elif array_type.startswith('Map'):
            array_type = 'HashMap'
        code_editor.add_line("%s = new %s[%s];" %
                             (value_name, array_type, ']['.join([str(dim) for dim in dims])))

        indexes = []
        for dim in dims:
            index = code_editor.new_tempvar()
            indexes.append(index)
            code_editor.add_line("for (int %s = 0; %s < %s; %s++)" %
                                 (index, index, dim, index))
            code_editor.increase_indentation()

        value_name += "[%s]" % ']['.join(indexes)
        self.gen_deserializer(data_name, offset_name, value_name, array_type_tree)
        for _ in range(len(dims)):
            code_editor.decrease_indentation()


    def _gen_deserializer_string(self, data_name, offset_name, value_name, tree):
        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen string
        self._code_editor.add_line("%s = new String(%s, %s, %s, Charset.forName(\"ISO-8859-1\"));" %
                                   (value_name, data_name, offset_name, size))
        self._code_editor.add_line("%s += %s;" % (offset_name, size))


    def _gen_deserializer_simples(self, data_name, offset_name, value_name, tree):
        value_type = self._variable_type_generator.gen_vartype(tree)
        if tree[0] in [Tokens.UnsignedByte, Tokens.Byte]:
            self._code_editor.add_line("%s = %s[%s];" %
                                       (value_name, data_name, offset_name or 0))
        elif tree[0] == Tokens.Boolean:
            self._code_editor.add_line("%s = (%s[%s] == 0) ? false : true;" %
                                       (value_name, data_name, offset_name or 0))
        elif tree[0] == Tokens.Char:
            self._code_editor.add_line("%s = (char) %s[%s];" %
                                       (value_name, data_name, offset_name or 0))
        else:
            method = self._simples_conversion_method_table[tree[0]]
            self._code_editor.add_line("%s = ByteBuffer.wrap(Arrays.copyOfRange(%s, %s, %s + %s.BYTES)).order(ByteOrder.LITTLE_ENDIAN).get%s();" %
                                        (value_name, data_name, offset_name or 0, offset_name or 0, value_type, method))
        if offset_name is not None:
            if tree[0] == Tokens.Boolean:
                value_type = 'Byte'
            self._code_editor.add_line("%s += %s.BYTES;" % (offset_name, value_type))


    def _gen_size(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        size_len = code_editor.new_tempvar()
        code_editor.add_line("byte %s;" % size_len)
        self._gen_deserializer_simples(data_name, offset_name, size_len, (Tokens.UnsignedByte, ))
        size_bytes = code_editor.new_tempvar()
        code_editor.add_line("byte[] %s = Arrays.copyOfRange(%s, %s, %s + %s);" %
                             (size_bytes, data_name, offset_name, offset_name, size_len))
        code_editor.add_line("%s += %s;" % (offset_name, size_len))

        size = code_editor.new_tempvar()
        code_editor.add_line("int %s;" % size)
        self._gen_deserializer_simples(size_bytes, None, size, (Tokens.UnsignedInteger, ))
        code_editor.add_line()

        return size


###################################################################################
###################################################################################


class VariableTypeGenerator:

    def __init__(self, user_defined_types):
        self._user_defined_types = user_defined_types

        self._simples_token_table = {
            Tokens.Boolean: ('Boolean', 'bool'),
            Tokens.Char: ('Character', 'char'),
            Tokens.Byte: ('Byte', 'byte'),
            Tokens.UnsignedByte: ('Byte', 'byte'),
            Tokens.Short: ('Short', 'short'),
            Tokens.UnsignedShort: ('Short', 'short'),
            Tokens.Integer: ('Integer', 'int'),
            Tokens.UnsignedInteger: ('Integer', 'int'),
            Tokens.Long: ('Long', 'long'),
            Tokens.UnsignedLong: ('Long', 'long'),
            Tokens.Float: ('Float', 'float'),
            Tokens.Double: ('Double', 'double'),
            Tokens.String: ('String', 'String')
        }


    def gen_vartype(self, tree, is_primitive=False):
        vartype = tree[0]

        for type_token in self._user_defined_types:
            if vartype in self._user_defined_types[type_token]:
                return vartype

        gen_vartype_func = getattr(
            self,
            '_gen_vartype_%s' % vartype,
            self._gen_vartype_simples
        )

        return gen_vartype_func(tree, is_primitive)


    def _gen_vartype_list(self, tree, is_primitive):
        return "List<%s>" % self.gen_vartype(tree[1])


    def _gen_vartype_map(self, tree, is_primitive):
        first_type = self.gen_vartype(tree[1][0])
        second_type = self.gen_vartype(tree[1][1])
        return "Map<%s, %s>" % (first_type, second_type)


    def _gen_vartype_array(self, tree, is_primitive):
        dims = list(reversed(tree[1][0]))
        initial_type = self.gen_vartype(tree[1][1])

        array_type = initial_type
        dim = '][' * (len(dims) - 1)
        array_type = "%s[%s]" % (array_type, dim)
        return array_type


    def _gen_vartype_simples(self, tree, is_primitive):
        primitive_index = 1 if is_primitive else 0
        return self._simples_token_table[tree[0]][primitive_index]



# -*- coding: utf-8 -*-

# python imports
import re
import inspect

# project imports
from .parser_core import Patterns, Rules


class RuleParser:

    def __init__(self):
        self._regex_factory = RegexFactory()
        self._current_type = None
        self._user_defined_types = None
        self._rules = Rules(self._regex_factory, self._get_type_tree)
        self._rule_funcs = self._find_rule_funcs()


    def get_type_tree(self, text, current_type, user_defined_types):

        def remove_white_spaces(text):
            return re.sub(
                "(\s*)?([\{\}\[\]\,\<\>\(\)\-])(\s*)?",
                lambda matchobj: matchobj.group(2),
                text
            )

        self._current_type = current_type
        self._user_defined_types = user_defined_types
        clear_text = remove_white_spaces(text)
        tree = self._get_type_tree(clear_text)
        return tree


    def _get_type_tree(self, text):
        for func in self._rule_funcs:
            tree = func(text)
            if tree is not None:
                return tree

        if text in self._user_defined_types and text != self._current_type:
            return (text, )

        raise Exception("error near '%s'" % text)


    def _find_rule_funcs(self):
        funcs = []
        for name, func in inspect.getmembers(self._rules, predicate=inspect.ismethod):
            if name.startswith('_rule_'):
                funcs.append(func)
        return funcs



class RegexFactory:

    def __init__(self):
        self._regex_table = {}
        self._load_patterns()


    def _load_patterns(self):
        for name, pattern in inspect.getmembers(Patterns):
            if not name.startswith('_'):
                self.add(name, pattern)


    def add(self, name, pattern):
        self._regex_table[name] = re.compile(pattern)


    def remove(self, name):
        del self._regex_table[name]


    def get(self, name):
        return self._regex_table[name]


    def match(self, name, text):
        regex = self.get(name)
        return regex.match(text)



# -*- coding: utf-8 -*-

# python imports
import os

# project imports
from ..parser_core import capitalization_rules
from .java_core import RootGenerator, HeaderGenerator, FooterGenerator, TypeGenerator
from .cpp import CppCodeEditor


class JavaCodeGenerator:

    def __init__(self):
        self._root_code_editor = JavaCodeEditor()
        self._root_generator = RootGenerator(self._root_code_editor)

        self._code_editor = JavaCodeEditor()
        self._header_generator = HeaderGenerator(self._code_editor)
        self._type_generator = TypeGenerator(self._code_editor)
        self._footer_generator = FooterGenerator(self._code_editor)

        self._generated_result = []


    def generate(self, parse_tree, module_name):
        result = []
        result += self._generate_root()
        result += self._generate(parse_tree, module_name)
        return result


    def _generate(self, parse_tree, module_name):
        result = []

        for user_defined_type, properties in parse_tree:
            self._header_generator.gen_header(module_name)
            self._type_generator.gen_type(user_defined_type, properties)
            self._footer_generator.gen_footer(module_name)

            code = self._code_editor.get_code()
            dirname = os.path.join('ks', module_name)
            filename = '%s.java' % capitalization_rules['PascalCase'](user_defined_type)
            result.append((code, os.path.join(dirname, filename)))
            self._code_editor.clear()

        return result


    def _generate_root(self):
        self._root_generator.gen_root()
        return [(self._root_code_editor.get_code(), 'ks/KSObject.java')]



class JavaCodeEditor(CppCodeEditor):

    pass



# -*- coding: utf-8 -*-

# project imports
from ..parser_core import Tokens


class HeaderGenerator:

    def __init__(self, code_editor):
        self._code_editor = code_editor


    def gen_header(self, module_name):
        code_editor = self._code_editor

        code_editor.add_line("#ifndef _KS_%s_H_" % module_name.upper())
        code_editor.add_line("#define _KS_%s_H_\n" % module_name.upper())
        self._gen_includes()
        code_editor.add_line('\n')
        self._gen_namespace('ks')
        code_editor.add_line()
        self._gen_ks_object()
        code_editor.add_line('\n')
        self._gen_namespace(module_name)
        code_editor.add_line()


    def _gen_includes(self):
        self._code_editor.add_line("#include <string>")
        self._code_editor.add_line("#include <vector>")
        self._code_editor.add_line("#include <map>")
        self._code_editor.add_line("#include <array>")


    def _gen_ks_object(self):
        self._code_editor.add_lines([
            "#ifndef _KS_OBJECT_",
            "#define _KS_OBJECT_",
            "",
            "class KSObject",
            "{",
            "public:",
            "\tstatic inline const std::string nameStatic() { return \"\"; }",
            "\tvirtual inline const std::string name() const = 0;",
            "\tvirtual std::string serialize() const = 0;",
            "\tvirtual unsigned int deserialize(const std::string &, unsigned int = 0) = 0;",
            "};",
            "",
            "#endif // _KS_OBJECT_",
        ])


    def _gen_namespace(self, ns_name):
        self._code_editor.add_line("namespace %s" % ns_name)
        self._code_editor.add_line('{')


###################################################################################
###################################################################################


class FooterGenerator:

    def __init__(self, code_editor):
        self._code_editor = code_editor


    def gen_footer(self, module_name):
        self._gen_namespace(module_name)
        self._code_editor.add_line()
        self._gen_namespace('ks')
        self._code_editor.add_line()
        self._code_editor.add_line("#endif // _KS_%s_H_" % module_name.upper())


    def _gen_namespace(self, ns_name):
        self._code_editor.add_line("} // namespace %s" % ns_name)


###################################################################################
###################################################################################


class TypeGenerator:

    def __init__(self, code_editor):
        self._user_defined_types = {
            Tokens.Class: {},
            Tokens.Enum: {}
        }

        self._code_editor = code_editor
        self._serializer_generator = SerializerGenerator(self._user_defined_types, code_editor)
        self._deserializer_generator = DeserializerGenerator(self._user_defined_types, code_editor)
        self._variable_type_generator = VariableTypeGenerator(self._user_defined_types)


    def gen_type(self, type_name, properties):
        gen_func = getattr(self, '_gen_type_%s' % properties['_def'][0])
        gen_func(type_name, properties)



    def _gen_type_class(self, type_name, properties):
        code_editor = self._code_editor

        _def = properties.pop('_def')
        properties = properties.items()

        parents = ['KSObject'] if len(_def) == 1 else [p[0] for p in _def[1]]
        self._user_defined_types[Tokens.Class][type_name] = (parents, properties)

        # generate definitions
        code_editor.add_line("class %s : public %s" % (type_name, ', '.join(parents)))
        code_editor.add_line("{\n")
        code_editor.add_line("protected:\n")
        for attr_name, tree in properties:
            attr_name = '__' + attr_name
            vartype = self._variable_type_generator.gen_vartype(tree)
            code_editor.add_line("%s %s;" % (vartype, attr_name), 1)

        code_editor.add_line()
        for attr_name, tree in properties:
            attr_name = '__has_' + attr_name
            code_editor.add_line("bool %s;" % attr_name, 1)

        code_editor.add_line('\n')

        # generate getters
        code_editor.add_line("public: // getters\n")
        code_editor.increase_indentation(put_accolade=False)
        for attr_name, tree in properties:
            vartype = self._variable_type_generator.gen_vartype(tree)
            code_editor.add_line("inline %s %s() const" % (vartype, attr_name))
            code_editor.increase_indentation()
            code_editor.add_line("return __%s;" % attr_name)
            code_editor.decrease_indentation()
            code_editor.add_line()
        code_editor.decrease_indentation(put_accolade=False)
        code_editor.add_line()

        # generate reference getters
        code_editor.add_line("public: // reference getters\n")
        code_editor.increase_indentation(put_accolade=False)
        for attr_name, tree in properties:
            vartype = self._variable_type_generator.gen_vartype(tree)
            code_editor.add_line("inline %s &ref_%s() const" % (vartype, attr_name))
            code_editor.increase_indentation()
            code_editor.add_line("return (%s&) __%s;" % (vartype, attr_name))
            code_editor.decrease_indentation()
            code_editor.add_line()
        code_editor.decrease_indentation(put_accolade=False)
        code_editor.add_line()

        # generate setters
        code_editor.add_line("public: // setters\n")
        code_editor.increase_indentation(put_accolade=False)
        for attr_name, tree in properties:
            vartype = self._variable_type_generator.gen_vartype(tree)
            code_editor.add_line("inline void %s(const %s &%s)" % (attr_name, vartype, attr_name))
            code_editor.increase_indentation()
            code_editor.add_line("__%s = %s;" % (attr_name, attr_name))
            code_editor.add_line("has_%s(true);" % attr_name)
            code_editor.decrease_indentation()
            code_editor.add_line()
        code_editor.decrease_indentation(put_accolade=False)
        code_editor.add_line()

        # generate has_attribute getters
        code_editor.add_line("public: // has_attribute getters\n")
        code_editor.increase_indentation(put_accolade=False)
        for attr_name, _ in properties:
            attr_name = 'has_' + attr_name
            code_editor.add_line("inline bool %s() const" % attr_name)
            code_editor.increase_indentation()
            code_editor.add_line("return __%s;" % attr_name)
            code_editor.decrease_indentation()
            code_editor.add_line()
        code_editor.decrease_indentation(put_accolade=False)
        code_editor.add_line()

        # generate has_attribute setters
        code_editor.add_line("public: // has_attribute setters\n")
        code_editor.increase_indentation(put_accolade=False)
        for attr_name, _ in properties:
            attr_name = 'has_' + attr_name
            code_editor.add_line("inline void %s(const bool &%s)" % (attr_name, attr_name))
            code_editor.increase_indentation()
            code_editor.add_line("__%s = %s;" % (attr_name, attr_name))
            code_editor.decrease_indentation()
            code_editor.add_line()
        code_editor.decrease_indentation(put_accolade=False)
        code_editor.add_line()


        code_editor.add_line("public:\n")
        code_editor.increase_indentation(put_accolade=False)

        # generate constructor
        code_editor.add_line("%s()" % type_name)
        code_editor.increase_indentation()
        for attr_name, _ in properties:
            code_editor.add_line("has_%s(false);" % attr_name)
        code_editor.decrease_indentation()
        code_editor.add_line()

        # generate name
        code_editor.add_line("static inline const std::string nameStatic()")
        code_editor.increase_indentation()
        code_editor.add_line("return \"%s\";" % type_name)
        code_editor.decrease_indentation()
        code_editor.add_line()

        code_editor.add_line("virtual inline const std::string name() const")
        code_editor.increase_indentation()
        code_editor.add_line("return \"%s\";" % type_name)
        code_editor.decrease_indentation()
        code_editor.add_line()

        # generate serializer
        result_name = 's'
        code_editor.add_line("std::string serialize() const")
        code_editor.increase_indentation()
        code_editor.add_line("std::string %s = \"\";" % result_name)
        code_editor.add_line()

        if parents[0] != 'KSObject':
            code_editor.add_line("// serialize parents")
            for parent in parents:
                code_editor.add_line("%s += %s::serialize();" % (result_name, parent))
            code_editor.add_line()

        for attr_name, tree in properties:
            code_editor.add_line("// serialize %s" % attr_name)
            attr_name = '__' + attr_name
            self._serializer_generator.gen_serializer(attr_name, result_name, tree, True)
            code_editor.add_line()

        code_editor.add_line("return %s;" % result_name)
        code_editor.decrease_indentation()
        code_editor.add_line()

        # generate deserializer
        data_name = 's'
        offset_name = 'offset'
        code_editor.add_line("unsigned int deserialize(const std::string &%s, unsigned int %s=0)" %
                             (data_name, offset_name))
        code_editor.increase_indentation()

        if parents[0] != 'KSObject':
            code_editor.add_line("// deserialize parents")
            for parent in parents:
                code_editor.add_line("%s = %s::deserialize(%s, %s);" % 
                                     (offset_name, parent, data_name, offset_name))
            code_editor.add_line()

        for attr_name, tree in properties:
            code_editor.add_line("// deserialize %s" % attr_name)
            attr_name = '__' + attr_name
            self._deserializer_generator.gen_deserializer(data_name, offset_name, attr_name, tree, True)
            code_editor.add_line()

        code_editor.add_line("return %s;" % offset_name)
        code_editor.decrease_indentation()

        code_editor.decrease_indentation(add_semi=True)


    def _gen_type_enum(self, type_name, properties):
        code_editor = self._code_editor

        _def = properties.pop('_def')
        value_type = _def[1][0]
        self._user_defined_types[Tokens.Enum][type_name] = value_type
        code_editor.add_line("enum class %s" % type_name)
        code_editor.increase_indentation()

        value = 0
        for attr in _def[1][1]:
            name = attr[0]
            if len(attr) > 1:
                value = attr[1]
            code_editor.add_line("%s = %s," % (name, value))
            value += 1

        code_editor.decrease_indentation(add_semi=True)

###################################################################################
###################################################################################


class SerializerGenerator:

    def __init__(self, user_defined_types, code_editor):
        self._simples_token_table = {
            Tokens.Boolean: 'bool',
            Tokens.Char: 'char',
            Tokens.Byte: 'char',
            Tokens.UnsignedByte: 'unsigned char',
            Tokens.Short: 'short',
            Tokens.UnsignedShort: 'unsigned short',
            Tokens.Integer: 'int',
            Tokens.UnsignedInteger: 'unsigned int',
            Tokens.Long: 'long long',
            Tokens.UnsignedLong: 'unsigned long long',
            Tokens.Float: 'float',
            Tokens.Double: 'double'
        }

        self._user_defined_types = user_defined_types
        self._code_editor = code_editor


    def gen_serializer(self, value_name, result_name, tree, check_null=False):
        value_type = tree[0]
        serializer_func = None

        for type_token in self._user_defined_types:
            if value_type in self._user_defined_types[type_token]:
                serializer_func = getattr(self, '_gen_serializer_%s' % type_token)
                break

        if serializer_func is None:
            serializer_func = getattr(
                self,
                '_gen_serializer_%s' % value_type,
                self._gen_serializer_simples
            )

        if check_null:
            self._code_editor.add_line("%s += __has_%s;" % 
                                       (result_name, value_name[2:]))
            self._code_editor.add_line("if (__has_%s)" % value_name[2:])
            self._code_editor.increase_indentation()
        else:
            self._code_editor.add_line("%s += '\\x01';" % result_name)

        serializer_func(value_name, result_name, tree)

        if check_null:
            self._code_editor.decrease_indentation()



    def _gen_serializer_class(self, value_name, result_name, tree):
        self._code_editor.add_line("%s += %s.serialize();" % (result_name, value_name))


    def _gen_serializer_enum(self, value_name, result_name, tree):
        value_type = self._user_defined_types[Tokens.Enum][tree[0]][0]
        value_type = self._simples_token_table[value_type]
        self._gen_serializer_simples(
            "(%s) %s" % (value_type, value_name),
            result_name,
            self._user_defined_types[Tokens.Enum][tree[0]]
        )


    def _gen_serializer_list(self, value_name, result_name, tree):
        code_editor = self._code_editor

        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen items
        item = code_editor.new_tempvar()
        code_editor.add_line("for (auto &%s : %s)" % (item, value_name))
        code_editor.increase_indentation()
        self.gen_serializer(item, result_name, tree[1])
        code_editor.decrease_indentation()


    def _gen_serializer_map(self, value_name, result_name, tree):
        code_editor = self._code_editor

        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen items
        pair = code_editor.new_tempvar()
        code_editor.add_line("for (auto &%s : %s)" % (pair, value_name))
        code_editor.increase_indentation()
        self.gen_serializer("%s.first" % pair, result_name, tree[1][0])
        code_editor.add_line()
        self.gen_serializer("%s.second" % pair, result_name, tree[1][1])
        code_editor.decrease_indentation()


    def _gen_serializer_array(self, value_name, result_name, tree):
        code_editor = self._code_editor

        dims = tree[1][0]
        array_type = tree[1][1]

        indexes = []
        for dim in dims:
            index = code_editor.new_tempvar()
            indexes.append(index)
            code_editor.add_line("for (unsigned int %s = 0; %s < %s; %s++)" %
                                 (index, index, dim, index))
            code_editor.increase_indentation()

        value_name += "[%s]" % ']['.join(indexes)
        self.gen_serializer(value_name, result_name, array_type)
        for _ in range(len(dims)):
            code_editor.decrease_indentation()


    def _gen_serializer_string(self, value_name, result_name, tree):
        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen string
        self._code_editor.add_line("%s += %s;" % (result_name, value_name))


    def _gen_serializer_simples(self, value_name, result_name, tree):
        code_editor = self._code_editor

        i = code_editor.new_tempvar()
        tmp = code_editor.new_tempvar()
        rc = code_editor.new_tempvar()

        value_type = self._simples_token_table[tree[0]]
        size = "sizeof(%s)" % value_type
        code_editor.add_line("%s %s = %s;" % (value_type, tmp, value_name))
        code_editor.add_line("auto %s = reinterpret_cast<char*>(&%s);" % (rc, tmp))
        self._code_editor.add_line("%s += std::string(%s, %s);" % (result_name, rc, size))


    def _gen_size(self, value_name, result_name, tree):
        code_editor = self._code_editor

        size = code_editor.new_tempvar()
        code_editor.add_line("std::string %s = \"\";" % size)
        self._gen_serializer_simples("%s.size()" % value_name, size, (Tokens.UnsignedInteger, ))
        code_editor.add_line("while (%s.size() && %s.back() == 0)" % (size, size))
        code_editor.add_line("%s.pop_back();" % size, 1)

        self._gen_serializer_simples("%s.size()" % size, result_name, (Tokens.UnsignedByte, ))
        code_editor.add_line("%s += %s;" % (result_name, size))
        code_editor.add_line()


###################################################################################
###################################################################################


class DeserializerGenerator:

    def __init__(self, user_defined_types, code_editor):
        self._simples_token_table = {
            Tokens.Boolean: 'bool',
            Tokens.Char: 'char',
            Tokens.Byte: 'char',
            Tokens.UnsignedByte: 'unsigned char',
            Tokens.Short: 'short',
            Tokens.UnsignedShort: 'unsigned short',
            Tokens.Integer: 'int',
            Tokens.UnsignedInteger: 'unsigned int',
            Tokens.Long: 'long long',
            Tokens.UnsignedLong: 'unsigned long long',
            Tokens.Float: 'float',
            Tokens.Double: 'double'
        }

        self._user_defined_types = user_defined_types
        self._code_editor = code_editor
        self._variable_type_generator = VariableTypeGenerator(self._user_defined_types)



    def gen_deserializer(self, data_name, offset_name, value_name, tree, check_null=False):
        value_type = tree[0]
        deserializer_func = None

        for type_token in self._user_defined_types:
            if value_type in self._user_defined_types[type_token]:
                deserializer_func = getattr(self, '_gen_deserializer_%s' % type_token)
                break

        if deserializer_func is None:
            deserializer_func = getattr(
                self,
                '_gen_deserializer_%s' % value_type,
                self._gen_deserializer_simples
            )

        if check_null:
            self._gen_deserializer_simples(data_name, offset_name, '__has_' + value_name[2:], 
                                           (Tokens.UnsignedByte, ))
            self._code_editor.add_line("if (__has_%s)" % value_name[2:])
            self._code_editor.increase_indentation()
        else:
            self._code_editor.add_line("%s++;" % offset_name)

        deserializer_func(data_name, offset_name, value_name, tree)

        if check_null:
            self._code_editor.decrease_indentation()



    def _gen_deserializer_class(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        code_editor.add_line("%s = %s.deserialize(%s, %s);" % 
                             (offset_name, value_name, data_name, offset_name))


    def _gen_deserializer_enum(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        enum_type = tree[0]
        val = code_editor.new_tempvar()
        val_type = self._user_defined_types[Tokens.Enum][tree[0]][0]
        val_type = self._simples_token_table[val_type]
        code_editor.add_line("%s %s;" % (val_type, val))

        self._gen_deserializer_simples(
            data_name,
            offset_name,
            val,
            self._user_defined_types[Tokens.Enum][tree[0]]
        )
        code_editor.add_line("%s = (%s) %s;" % (value_name, enum_type, val))


    def _gen_deserializer_list(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen items
        code_editor.add_line("%s.clear();" % value_name)
        i = code_editor.new_tempvar()
        code_editor.add_line("for (unsigned int %s = 0; %s < %s; %s++)" % (i, i, size, i))
        code_editor.increase_indentation()
        item = code_editor.new_tempvar()
        item_type = self._variable_type_generator.gen_vartype(tree[1])
        code_editor.add_line("%s %s;" % (item_type, item))
        self.gen_deserializer(data_name, offset_name, item, tree[1])
        code_editor.add_line("%s.push_back(%s);" % (value_name, item))
        code_editor.decrease_indentation()


    def _gen_deserializer_map(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen items
        code_editor.add_line("%s.clear();" % value_name)
        i = code_editor.new_tempvar()
        code_editor.add_line("for (unsigned int %s = 0; %s < %s; %s++)" % (i, i, size, i))
        code_editor.increase_indentation()

        key = code_editor.new_tempvar()
        val = code_editor.new_tempvar()
        key_type = self._variable_type_generator.gen_vartype(tree[1][0])
        val_type = self._variable_type_generator.gen_vartype(tree[1][1])

        code_editor.add_line("%s %s;" % (key_type, key))
        self.gen_deserializer(data_name, offset_name, key, tree[1][0])
        code_editor.add_line()

        code_editor.add_line("%s %s;" % (val_type, val))
        self.gen_deserializer(data_name, offset_name, val, tree[1][1])
        code_editor.add_line()

        code_editor.add_line("%s[%s] = %s;" % (value_name, key, val))
        code_editor.decrease_indentation()


    def _gen_deserializer_array(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        dims = tree[1][0]
        array_type = tree[1][1]

        indexes = []
        for dim in dims:
            index = code_editor.new_tempvar()
            indexes.append(index)
            code_editor.add_line("for (unsigned int %s = 0; %s < %s; %s++)" %
                                 (index, index, dim, index))
            code_editor.increase_indentation()

        value_name += "[%s]" % ']['.join(indexes)
        self.gen_deserializer(data_name, offset_name, value_name, array_type)
        for _ in range(len(dims)):
            code_editor.decrease_indentation()


    def _gen_deserializer_string(self, data_name, offset_name, value_name, tree):
        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen string
        self._code_editor.add_line("%s = %s.substr(%s, %s);" %
                                   (value_name, data_name, offset_name, size))
        self._code_editor.add_line("%s += %s;" % (offset_name, size))


    def _gen_deserializer_simples(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor
        value_type = self._simples_token_table[tree[0]]
        code_editor.add_line("%s = *((%s*) (&%s[%s]));" %
                             (value_name, value_type, data_name, offset_name or 0))
        if offset_name is not None:
            code_editor.add_line("%s += sizeof(%s);" % (offset_name, value_type))


    def _gen_size(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        size_len = code_editor.new_tempvar()
        code_editor.add_line("unsigned char %s;" % size_len)
        self._gen_deserializer_simples(data_name, offset_name, size_len, (Tokens.UnsignedByte, ))
        size_bytes = code_editor.new_tempvar()
        code_editor.add_line("std::string %s = std::string(&%s[%s], %s);" %
                              (size_bytes, data_name, offset_name, size_len))
        code_editor.add_line("%s += %s;" % (offset_name, size_len))

        code_editor.add_line("while (%s.size() < sizeof(unsigned int))" % size_bytes)
        code_editor.add_line("%s += '\\x00';" % size_bytes, 1)

        size = code_editor.new_tempvar()
        code_editor.add_line("unsigned int %s;" % size)
        self._gen_deserializer_simples(size_bytes, None, size, (Tokens.UnsignedInteger, ))
        code_editor.add_line()

        return size


###################################################################################
###################################################################################


class VariableTypeGenerator:

    def __init__(self, user_defined_types):
        self._user_defined_types = user_defined_types

        self._simples_token_table = {
            Tokens.Boolean: 'bool',
            Tokens.Char: 'char',
            Tokens.Byte: 'char',
            Tokens.UnsignedByte: 'unsigned char',
            Tokens.Short: 'short',
            Tokens.UnsignedShort: 'unsigned short',
            Tokens.Integer: 'int',
            Tokens.UnsignedInteger: 'unsigned int',
            Tokens.Long: 'long long',
            Tokens.UnsignedLong: 'unsigned long long',
            Tokens.Float: 'float',
            Tokens.Double: 'double',
            Tokens.String: 'std::string'
        }


    def gen_vartype(self, tree):
        vartype = tree[0]

        for type_token in self._user_defined_types:
            if vartype in self._user_defined_types[type_token]:
                return vartype

        gen_vartype_func = getattr(
            self,
            '_gen_vartype_%s' % vartype,
            self._gen_vartype_simples
        )

        return gen_vartype_func(tree)


    def _gen_vartype_list(self, tree):
        return "std::vector<%s>" % self.gen_vartype(tree[1])


    def _gen_vartype_map(self, tree):
        first_type = self.gen_vartype(tree[1][0])
        second_type = self.gen_vartype(tree[1][1])
        return "std::map<%s, %s>" % (first_type, second_type)


    def _gen_vartype_array(self, tree):
        dims = list(reversed(tree[1][0]))
        initial_type = self.gen_vartype(tree[1][1])

        array_type = initial_type
        for dim in dims:
            array_type = "std::array<%s, %s>" % (array_type, dim)
        return array_type


    def _gen_vartype_simples(self, tree):
        return self._simples_token_table[tree[0]]



# -*- coding: utf-8 -*-

# python imports
import sys
import struct
from enum import Enum

PY3 = sys.version_info > (3,)


class EColor(Enum):
	White = 0
	Red = 3
	Green = 4
	Blue = -2
	Black = -1


class Parent(object):

	@staticmethod
	def name():
		return 'Parent'


	def __init__(self, first_name=None, _last_name_=None):
		self.initialize(first_name, _last_name_)
	

	def initialize(self, first_name=None, _last_name_=None):
		self.first_name = first_name
		self._last_name_ = _last_name_
	

	def serialize(self):
		s = b''
		
		# serialize self.first_name
		s += b'\x00' if self.first_name is None else b'\x01'
		if self.first_name is not None:
			tmp0 = b''
			tmp0 += struct.pack('I', len(self.first_name))
			while len(tmp0) and tmp0[-1] == b'\x00'[0]:
				tmp0 = tmp0[:-1]
			s += struct.pack('B', len(tmp0))
			s += tmp0
			
			s += self.first_name.encode('ISO-8859-1') if PY3 else self.first_name
		
		# serialize self._last_name_
		s += b'\x00' if self._last_name_ is None else b'\x01'
		if self._last_name_ is not None:
			tmp1 = b''
			tmp1 += struct.pack('I', len(self._last_name_))
			while len(tmp1) and tmp1[-1] == b'\x00'[0]:
				tmp1 = tmp1[:-1]
			s += struct.pack('B', len(tmp1))
			s += tmp1
			
			s += self._last_name_.encode('ISO-8859-1') if PY3 else self._last_name_
		
		return s
	

	def deserialize(self, s, offset=0):
		# deserialize self.first_name
		tmp2 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp2:
			tmp3 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp4 = s[offset:offset + tmp3]
			offset += tmp3
			tmp4 += b'\x00' * (4 - tmp3)
			tmp5 = struct.unpack('I', tmp4)[0]
			
			self.first_name = s[offset:offset + tmp5].decode('ISO-8859-1') if PY3 else s[offset:offset + tmp5]
			offset += tmp5
		else:
			self.first_name = None
		
		# deserialize self._last_name_
		tmp6 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp6:
			tmp7 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp8 = s[offset:offset + tmp7]
			offset += tmp7
			tmp8 += b'\x00' * (4 - tmp7)
			tmp9 = struct.unpack('I', tmp8)[0]
			
			self._last_name_ = s[offset:offset + tmp9].decode('ISO-8859-1') if PY3 else s[offset:offset + tmp9]
			offset += tmp9
		else:
			self._last_name_ = None
		
		return offset


class Child(Parent):

	@staticmethod
	def name():
		return 'Child'


	def __init__(self, first_name=None, _last_name_=None, c=None):
		self.initialize(first_name, _last_name_, c)
	

	def initialize(self, first_name=None, _last_name_=None, c=None):
		Parent.initialize(self, first_name, _last_name_)
		
		self.c = c
	

	def serialize(self):
		s = b''
		
		# serialize parents
		s += Parent.serialize(self)
		
		# serialize self.c
		s += b'\x00' if self.c is None else b'\x01'
		if self.c is not None:
			tmp10 = b''
			tmp10 += struct.pack('I', len(self.c))
			while len(tmp10) and tmp10[-1] == b'\x00'[0]:
				tmp10 = tmp10[:-1]
			s += struct.pack('B', len(tmp10))
			s += tmp10
			
			s += self.c.encode('ISO-8859-1') if PY3 else self.c
		
		return s
	

	def deserialize(self, s, offset=0):
		# deserialize parents
		offset = Parent.deserialize(self, s, offset)
		
		# deserialize self.c
		tmp11 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp11:
			tmp12 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp13 = s[offset:offset + tmp12]
			offset += tmp12
			tmp13 += b'\x00' * (4 - tmp12)
			tmp14 = struct.unpack('I', tmp13)[0]
			
			self.c = s[offset:offset + tmp14].decode('ISO-8859-1') if PY3 else s[offset:offset + tmp14]
			offset += tmp14
		else:
			self.c = None
		
		return offset


class Test(object):

	@staticmethod
	def name():
		return 'Test'


	def __init__(self, v0=None, v1=None, v2=None, v3=None, v4=None, v5=None, v6=None, v7=None, v8=None, v9=None, v10=None, v11=None, v12=None, v13=None, v14=None, v15=None, v16=None, v17=None, v18=None, v19=None, v20=None, v21=None, v22=None, v23=None):
		self.initialize(v0, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10, v11, v12, v13, v14, v15, v16, v17, v18, v19, v20, v21, v22, v23)
	

	def initialize(self, v0=None, v1=None, v2=None, v3=None, v4=None, v5=None, v6=None, v7=None, v8=None, v9=None, v10=None, v11=None, v12=None, v13=None, v14=None, v15=None, v16=None, v17=None, v18=None, v19=None, v20=None, v21=None, v22=None, v23=None):
		self.v0 = v0
		self.v1 = v1
		self.v2 = v2
		self.v3 = v3
		self.v4 = v4
		self.v5 = v5
		self.v6 = v6
		self.v7 = v7
		self.v8 = v8
		self.v9 = v9
		self.v10 = v10
		self.v11 = v11
		self.v12 = v12
		self.v13 = v13
		self.v14 = v14
		self.v15 = v15
		self.v16 = v16
		self.v17 = v17
		self.v18 = v18
		self.v19 = v19
		self.v20 = v20
		self.v21 = v21
		self.v22 = v22
		self.v23 = v23
	

	def serialize(self):
		s = b''
		
		# serialize self.v0
		s += b'\x00' if self.v0 is None else b'\x01'
		if self.v0 is not None:
			s += struct.pack('?', self.v0)
		
		# serialize self.v1
		s += b'\x00' if self.v1 is None else b'\x01'
		if self.v1 is not None:
			s += struct.pack('c', self.v1.encode('ISO-8859-1') if PY3 else self.v1)
		
		# serialize self.v2
		s += b'\x00' if self.v2 is None else b'\x01'
		if self.v2 is not None:
			s += struct.pack('b', self.v2)
		
		# serialize self.v3
		s += b'\x00' if self.v3 is None else b'\x01'
		if self.v3 is not None:
			s += struct.pack('B', self.v3)
		
		# serialize self.v4
		s += b'\x00' if self.v4 is None else b'\x01'
		if self.v4 is not None:
			s += struct.pack('h', self.v4)
		
		# serialize self.v5
		s += b'\x00' if self.v5 is None else b'\x01'
		if self.v5 is not None:
			s += struct.pack('H', self.v5)
		
		# serialize self.v6
		s += b'\x00' if self.v6 is None else b'\x01'
		if self.v6 is not None:
			s += struct.pack('i', self.v6)
		
		# serialize self.v7
		s += b'\x00' if self.v7 is None else b'\x01'
		if self.v7 is not None:
			s += struct.pack('I', self.v7)
		
		# serialize self.v8
		s += b'\x00' if self.v8 is None else b'\x01'
		if self.v8 is not None:
			s += struct.pack('q', self.v8)
		
		# serialize self.v9
		s += b'\x00' if self.v9 is None else b'\x01'
		if self.v9 is not None:
			s += struct.pack('Q', self.v9)
		
		# serialize self.v10
		s += b'\x00' if self.v10 is None else b'\x01'
		if self.v10 is not None:
			s += struct.pack('f', self.v10)
		
		# serialize self.v11
		s += b'\x00' if self.v11 is None else b'\x01'
		if self.v11 is not None:
			s += struct.pack('d', self.v11)
		
		# serialize self.v12
		s += b'\x00' if self.v12 is None else b'\x01'
		if self.v12 is not None:
			tmp15 = b''
			tmp15 += struct.pack('I', len(self.v12))
			while len(tmp15) and tmp15[-1] == b'\x00'[0]:
				tmp15 = tmp15[:-1]
			s += struct.pack('B', len(tmp15))
			s += tmp15
			
			s += self.v12.encode('ISO-8859-1') if PY3 else self.v12
		
		# serialize self.v13
		s += b'\x00' if self.v13 is None else b'\x01'
		if self.v13 is not None:
			s += struct.pack('b', self.v13.value)
		
		# serialize self.v14
		s += b'\x00' if self.v14 is None else b'\x01'
		if self.v14 is not None:
			s += self.v14.serialize()
		
		# serialize self.v15
		s += b'\x00' if self.v15 is None else b'\x01'
		if self.v15 is not None:
			tmp16 = b''
			tmp16 += struct.pack('I', len(self.v15))
			while len(tmp16) and tmp16[-1] == b'\x00'[0]:
				tmp16 = tmp16[:-1]
			s += struct.pack('B', len(tmp16))
			s += tmp16
			
			for tmp17 in self.v15:
				s += b'\x00' if tmp17 is None else b'\x01'
				if tmp17 is not None:
					s += struct.pack('i', tmp17)
		
		# serialize self.v16
		s += b'\x00' if self.v16 is None else b'\x01'
		if self.v16 is not None:
			tmp18 = b''
			tmp18 += struct.pack('I', len(self.v16))
			while len(tmp18) and tmp18[-1] == b'\x00'[0]:
				tmp18 = tmp18[:-1]
			s += struct.pack('B', len(tmp18))
			s += tmp18
			
			for tmp19 in self.v16:
				s += b'\x00' if tmp19 is None else b'\x01'
				if tmp19 is not None:
					tmp20 = b''
					tmp20 += struct.pack('I', len(tmp19))
					while len(tmp20) and tmp20[-1] == b'\x00'[0]:
						tmp20 = tmp20[:-1]
					s += struct.pack('B', len(tmp20))
					s += tmp20
					
					for tmp21 in tmp19:
						s += b'\x00' if tmp21 is None else b'\x01'
						if tmp21 is not None:
							s += struct.pack('c', tmp21.encode('ISO-8859-1') if PY3 else tmp21)
		
		# serialize self.v17
		s += b'\x00' if self.v17 is None else b'\x01'
		if self.v17 is not None:
			tmp22 = b''
			tmp22 += struct.pack('I', len(self.v17))
			while len(tmp22) and tmp22[-1] == b'\x00'[0]:
				tmp22 = tmp22[:-1]
			s += struct.pack('B', len(tmp22))
			s += tmp22
			
			for tmp23 in self.v17:
				s += b'\x00' if tmp23 is None else b'\x01'
				if tmp23 is not None:
					tmp24 = b''
					tmp24 += struct.pack('I', len(tmp23))
					while len(tmp24) and tmp24[-1] == b'\x00'[0]:
						tmp24 = tmp24[:-1]
					s += struct.pack('B', len(tmp24))
					s += tmp24
					
					s += tmp23.encode('ISO-8859-1') if PY3 else tmp23
				s += b'\x00' if self.v17[tmp23] is None else b'\x01'
				if self.v17[tmp23] is not None:
					s += struct.pack('i', self.v17[tmp23])
		
		# serialize self.v18
		s += b'\x00' if self.v18 is None else b'\x01'
		if self.v18 is not None:
			tmp25 = b''
			tmp25 += struct.pack('I', len(self.v18))
			while len(tmp25) and tmp25[-1] == b'\x00'[0]:
				tmp25 = tmp25[:-1]
			s += struct.pack('B', len(tmp25))
			s += tmp25
			
			for tmp26 in self.v18:
				s += b'\x00' if tmp26 is None else b'\x01'
				if tmp26 is not None:
					s += struct.pack('c', tmp26.encode('ISO-8859-1') if PY3 else tmp26)
				s += b'\x00' if self.v18[tmp26] is None else b'\x01'
				if self.v18[tmp26] is not None:
					tmp27 = b''
					tmp27 += struct.pack('I', len(self.v18[tmp26]))
					while len(tmp27) and tmp27[-1] == b'\x00'[0]:
						tmp27 = tmp27[:-1]
					s += struct.pack('B', len(tmp27))
					s += tmp27
					
					for tmp28 in self.v18[tmp26]:
						s += b'\x00' if tmp28 is None else b'\x01'
						if tmp28 is not None:
							tmp29 = b''
							tmp29 += struct.pack('I', len(tmp28))
							while len(tmp29) and tmp29[-1] == b'\x00'[0]:
								tmp29 = tmp29[:-1]
							s += struct.pack('B', len(tmp29))
							s += tmp29
							
							for tmp30 in tmp28:
								s += b'\x00' if tmp30 is None else b'\x01'
								if tmp30 is not None:
									s += struct.pack('d', tmp30)
								s += b'\x00' if tmp28[tmp30] is None else b'\x01'
								if tmp28[tmp30] is not None:
									s += struct.pack('b', tmp28[tmp30].value)
		
		# serialize self.v19
		s += b'\x00' if self.v19 is None else b'\x01'
		if self.v19 is not None:
			for tmp31 in range(10):
				s += b'\x00' if self.v19[tmp31] is None else b'\x01'
				if self.v19[tmp31] is not None:
					s += struct.pack('b', self.v19[tmp31])
		
		# serialize self.v20
		s += b'\x00' if self.v20 is None else b'\x01'
		if self.v20 is not None:
			for tmp32 in range(10):
				for tmp33 in range(20):
					s += b'\x00' if self.v20[tmp32][tmp33] is None else b'\x01'
					if self.v20[tmp32][tmp33] is not None:
						tmp34 = b''
						tmp34 += struct.pack('I', len(self.v20[tmp32][tmp33]))
						while len(tmp34) and tmp34[-1] == b'\x00'[0]:
							tmp34 = tmp34[:-1]
						s += struct.pack('B', len(tmp34))
						s += tmp34
						
						for tmp35 in self.v20[tmp32][tmp33]:
							s += b'\x00' if tmp35 is None else b'\x01'
							if tmp35 is not None:
								tmp36 = b''
								tmp36 += struct.pack('I', len(tmp35))
								while len(tmp36) and tmp36[-1] == b'\x00'[0]:
									tmp36 = tmp36[:-1]
								s += struct.pack('B', len(tmp36))
								s += tmp36
								
								s += tmp35.encode('ISO-8859-1') if PY3 else tmp35
		
		# serialize self.v21
		s += b'\x00' if self.v21 is None else b'\x01'
		if self.v21 is not None:
			tmp37 = b''
			tmp37 += struct.pack('I', len(self.v21))
			while len(tmp37) and tmp37[-1] == b'\x00'[0]:
				tmp37 = tmp37[:-1]
			s += struct.pack('B', len(tmp37))
			s += tmp37
			
			for tmp38 in self.v21:
				s += b'\x00' if tmp38 is None else b'\x01'
				if tmp38 is not None:
					for tmp39 in range(4):
						s += b'\x00' if tmp38[tmp39] is None else b'\x01'
						if tmp38[tmp39] is not None:
							s += tmp38[tmp39].serialize()
		
		# serialize self.v22
		s += b'\x00' if self.v22 is None else b'\x01'
		if self.v22 is not None:
			tmp40 = b''
			tmp40 += struct.pack('I', len(self.v22))
			while len(tmp40) and tmp40[-1] == b'\x00'[0]:
				tmp40 = tmp40[:-1]
			s += struct.pack('B', len(tmp40))
			s += tmp40
			
			for tmp41 in self.v22:
				s += b'\x00' if tmp41 is None else b'\x01'
				if tmp41 is not None:
					tmp42 = b''
					tmp42 += struct.pack('I', len(tmp41))
					while len(tmp42) and tmp42[-1] == b'\x00'[0]:
						tmp42 = tmp42[:-1]
					s += struct.pack('B', len(tmp42))
					s += tmp42
					
					s += tmp41.encode('ISO-8859-1') if PY3 else tmp41
				s += b'\x00' if self.v22[tmp41] is None else b'\x01'
				if self.v22[tmp41] is not None:
					s += self.v22[tmp41].serialize()
		
		# serialize self.v23
		s += b'\x00' if self.v23 is None else b'\x01'
		if self.v23 is not None:
			for tmp43 in range(5):
				for tmp44 in range(10):
					s += b'\x00' if self.v23[tmp43][tmp44] is None else b'\x01'
					if self.v23[tmp43][tmp44] is not None:
						s += self.v23[tmp43][tmp44].serialize()
		
		return s
	

	def deserialize(self, s, offset=0):
		# deserialize self.v0
		tmp45 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp45:
			self.v0 = struct.unpack('?', s[offset:offset + 1])[0]
			offset += 1
		else:
			self.v0 = None
		
		# deserialize self.v1
		tmp46 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp46:
			self.v1 = struct.unpack('c', s[offset:offset + 1])[0]
			offset += 1
			self.v1 = self.v1.decode('ISO-8859-1') if PY3 else self.v1
		else:
			self.v1 = None
		
		# deserialize self.v2
		tmp47 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp47:
			self.v2 = struct.unpack('b', s[offset:offset + 1])[0]
			offset += 1
		else:
			self.v2 = None
		
		# deserialize self.v3
		tmp48 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp48:
			self.v3 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
		else:
			self.v3 = None
		
		# deserialize self.v4
		tmp49 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp49:
			self.v4 = struct.unpack('h', s[offset:offset + 2])[0]
			offset += 2
		else:
			self.v4 = None
		
		# deserialize self.v5
		tmp50 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp50:
			self.v5 = struct.unpack('H', s[offset:offset + 2])[0]
			offset += 2
		else:
			self.v5 = None
		
		# deserialize self.v6
		tmp51 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp51:
			self.v6 = struct.unpack('i', s[offset:offset + 4])[0]
			offset += 4
		else:
			self.v6 = None
		
		# deserialize self.v7
		tmp52 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp52:
			self.v7 = struct.unpack('I', s[offset:offset + 4])[0]
			offset += 4
		else:
			self.v7 = None
		
		# deserialize self.v8
		tmp53 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp53:
			self.v8 = struct.unpack('q', s[offset:offset + 8])[0]
			offset += 8
		else:
			self.v8 = None
		
		# deserialize self.v9
		tmp54 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp54:
			self.v9 = struct.unpack('Q', s[offset:offset + 8])[0]
			offset += 8
		else:
			self.v9 = None
		
		# deserialize self.v10
		tmp55 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp55:
			self.v10 = struct.unpack('f', s[offset:offset + 4])[0]
			offset += 4
		else:
			self.v10 = None
		
		# deserialize self.v11
		tmp56 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp56:
			self.v11 = struct.unpack('d', s[offset:offset + 8])[0]
			offset += 8
		else:
			self.v11 = None
		
		# deserialize self.v12
		tmp57 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp57:
			tmp58 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp59 = s[offset:offset + tmp58]
			offset += tmp58
			tmp59 += b'\x00' * (4 - tmp58)
			tmp60 = struct.unpack('I', tmp59)[0]
			
			self.v12 = s[offset:offset + tmp60].decode('ISO-8859-1') if PY3 else s[offset:offset + tmp60]
			offset += tmp60
		else:
			self.v12 = None
		
		# deserialize self.v13
		tmp61 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp61:
			tmp62 = struct.unpack('b', s[offset:offset + 1])[0]
			offset += 1
			self.v13 = EColor(tmp62)
		else:
			self.v13 = None
		
		# deserialize self.v14
		tmp63 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp63:
			self.v14 = Child()
			offset = self.v14.deserialize(s, offset)
		else:
			self.v14 = None
		
		# deserialize self.v15
		tmp64 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp64:
			tmp65 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp66 = s[offset:offset + tmp65]
			offset += tmp65
			tmp66 += b'\x00' * (4 - tmp65)
			tmp67 = struct.unpack('I', tmp66)[0]
			
			self.v15 = []
			for tmp68 in range(tmp67):
				tmp70 = struct.unpack('B', s[offset:offset + 1])[0]
				offset += 1
				if tmp70:
					tmp69 = struct.unpack('i', s[offset:offset + 4])[0]
					offset += 4
				else:
					tmp69 = None
				self.v15.append(tmp69)
		else:
			self.v15 = None
		
		# deserialize self.v16
		tmp71 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp71:
			tmp72 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp73 = s[offset:offset + tmp72]
			offset += tmp72
			tmp73 += b'\x00' * (4 - tmp72)
			tmp74 = struct.unpack('I', tmp73)[0]
			
			self.v16 = []
			for tmp75 in range(tmp74):
				tmp77 = struct.unpack('B', s[offset:offset + 1])[0]
				offset += 1
				if tmp77:
					tmp78 = struct.unpack('B', s[offset:offset + 1])[0]
					offset += 1
					tmp79 = s[offset:offset + tmp78]
					offset += tmp78
					tmp79 += b'\x00' * (4 - tmp78)
					tmp80 = struct.unpack('I', tmp79)[0]
					
					tmp76 = []
					for tmp81 in range(tmp80):
						tmp83 = struct.unpack('B', s[offset:offset + 1])[0]
						offset += 1
						if tmp83:
							tmp82 = struct.unpack('c', s[offset:offset + 1])[0]
							offset += 1
							tmp82 = tmp82.decode('ISO-8859-1') if PY3 else tmp82
						else:
							tmp82 = None
						tmp76.append(tmp82)
				else:
					tmp76 = None
				self.v16.append(tmp76)
		else:
			self.v16 = None
		
		# deserialize self.v17
		tmp84 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp84:
			tmp85 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp86 = s[offset:offset + tmp85]
			offset += tmp85
			tmp86 += b'\x00' * (4 - tmp85)
			tmp87 = struct.unpack('I', tmp86)[0]
			
			self.v17 = {}
			for tmp88 in range(tmp87):
				tmp91 = struct.unpack('B', s[offset:offset + 1])[0]
				offset += 1
				if tmp91:
					tmp92 = struct.unpack('B', s[offset:offset + 1])[0]
					offset += 1
					tmp93 = s[offset:offset + tmp92]
					offset += tmp92
					tmp93 += b'\x00' * (4 - tmp92)
					tmp94 = struct.unpack('I', tmp93)[0]
					
					tmp89 = s[offset:offset + tmp94].decode('ISO-8859-1') if PY3 else s[offset:offset + tmp94]
					offset += tmp94
				else:
					tmp89 = None
				tmp95 = struct.unpack('B', s[offset:offset + 1])[0]
				offset += 1
				if tmp95:
					tmp90 = struct.unpack('i', s[offset:offset + 4])[0]
					offset += 4
				else:
					tmp90 = None
				self.v17[tmp89] = tmp90
		else:
			self.v17 = None
		
		# deserialize self.v18
		tmp96 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp96:
			tmp97 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp98 = s[offset:offset + tmp97]
			offset += tmp97
			tmp98 += b'\x00' * (4 - tmp97)
			tmp99 = struct.unpack('I', tmp98)[0]
			
			self.v18 = {}
			for tmp100 in range(tmp99):
				tmp103 = struct.unpack('B', s[offset:offset + 1])[0]
				offset += 1
				if tmp103:
					tmp101 = struct.unpack('c', s[offset:offset + 1])[0]
					offset += 1
					tmp101 = tmp101.decode('ISO-8859-1') if PY3 else tmp101
				else:
					tmp101 = None
				tmp104 = struct.unpack('B', s[offset:offset + 1])[0]
				offset += 1
				if tmp104:
					tmp105 = struct.unpack('B', s[offset:offset + 1])[0]
					offset += 1
					tmp106 = s[offset:offset + tmp105]
					offset += tmp105
					tmp106 += b'\x00' * (4 - tmp105)
					tmp107 = struct.unpack('I', tmp106)[0]
					
					tmp102 = []
					for tmp108 in range(tmp107):
						tmp110 = struct.unpack('B', s[offset:offset + 1])[0]
						offset += 1
						if tmp110:
							tmp111 = struct.unpack('B', s[offset:offset + 1])[0]
							offset += 1
							tmp112 = s[offset:offset + tmp111]
							offset += tmp111
							tmp112 += b'\x00' * (4 - tmp111)
							tmp113 = struct.unpack('I', tmp112)[0]
							
							tmp109 = {}
							for tmp114 in range(tmp113):
								tmp117 = struct.unpack('B', s[offset:offset + 1])[0]
								offset += 1
								if tmp117:
									tmp115 = struct.unpack('d', s[offset:offset + 8])[0]
									offset += 8
								else:
									tmp115 = None
								tmp118 = struct.unpack('B', s[offset:offset + 1])[0]
								offset += 1
								if tmp118:
									tmp119 = struct.unpack('b', s[offset:offset + 1])[0]
									offset += 1
									tmp116 = EColor(tmp119)
								else:
									tmp116 = None
								tmp109[tmp115] = tmp116
						else:
							tmp109 = None
						tmp102.append(tmp109)
				else:
					tmp102 = None
				self.v18[tmp101] = tmp102
		else:
			self.v18 = None
		
		# deserialize self.v19
		tmp120 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp120:
			self.v19 = [None for _ in range(10)]
			for tmp121 in range(10):
				tmp122 = struct.unpack('B', s[offset:offset + 1])[0]
				offset += 1
				if tmp122:
					self.v19[tmp121] = struct.unpack('b', s[offset:offset + 1])[0]
					offset += 1
				else:
					self.v19[tmp121] = None
		else:
			self.v19 = None
		
		# deserialize self.v20
		tmp123 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp123:
			self.v20 = [[None for _ in range(10)] for _ in range(20)]
			for tmp124 in range(10):
				for tmp125 in range(20):
					tmp126 = struct.unpack('B', s[offset:offset + 1])[0]
					offset += 1
					if tmp126:
						tmp127 = struct.unpack('B', s[offset:offset + 1])[0]
						offset += 1
						tmp128 = s[offset:offset + tmp127]
						offset += tmp127
						tmp128 += b'\x00' * (4 - tmp127)
						tmp129 = struct.unpack('I', tmp128)[0]
						
						self.v20[tmp124][tmp125] = []
						for tmp130 in range(tmp129):
							tmp132 = struct.unpack('B', s[offset:offset + 1])[0]
							offset += 1
							if tmp132:
								tmp133 = struct.unpack('B', s[offset:offset + 1])[0]
								offset += 1
								tmp134 = s[offset:offset + tmp133]
								offset += tmp133
								tmp134 += b'\x00' * (4 - tmp133)
								tmp135 = struct.unpack('I', tmp134)[0]
								
								tmp131 = s[offset:offset + tmp135].decode('ISO-8859-1') if PY3 else s[offset:offset + tmp135]
								offset += tmp135
							else:
								tmp131 = None
							self.v20[tmp124][tmp125].append(tmp131)
					else:
						self.v20[tmp124][tmp125] = None
		else:
			self.v20 = None
		
		# deserialize self.v21
		tmp136 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp136:
			tmp137 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp138 = s[offset:offset + tmp137]
			offset += tmp137
			tmp138 += b'\x00' * (4 - tmp137)
			tmp139 = struct.unpack('I', tmp138)[0]
			
			self.v21 = []
			for tmp140 in range(tmp139):
				tmp142 = struct.unpack('B', s[offset:offset + 1])[0]
				offset += 1
				if tmp142:
					tmp141 = [None for _ in range(4)]
					for tmp143 in range(4):
						tmp144 = struct.unpack('B', s[offset:offset + 1])[0]
						offset += 1
						if tmp144:
							tmp141[tmp143] = Child()
							offset = tmp141[tmp143].deserialize(s, offset)
						else:
							tmp141[tmp143] = None
				else:
					tmp141 = None
				self.v21.append(tmp141)
		else:
			self.v21 = None
		
		# deserialize self.v22
		tmp145 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp145:
			tmp146 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp147 = s[offset:offset + tmp146]
			offset += tmp146
			tmp147 += b'\x00' * (4 - tmp146)
			tmp148 = struct.unpack('I', tmp147)[0]
			
			self.v22 = {}
			for tmp149 in range(tmp148):
				tmp152 = struct.unpack('B', s[offset:offset + 1])[0]
				offset += 1
				if tmp152:
					tmp153 = struct.unpack('B', s[offset:offset + 1])[0]
					offset += 1
					tmp154 = s[offset:offset + tmp153]
					offset += tmp153
					tmp154 += b'\x00' * (4 - tmp153)
					tmp155 = struct.unpack('I', tmp154)[0]
					
					tmp150 = s[offset:offset + tmp155].decode('ISO-8859-1') if PY3 else s[offset:offset + tmp155]
					offset += tmp155
				else:
					tmp150 = None
				tmp156 = struct.unpack('B', s[offset:offset + 1])[0]
				offset += 1
				if tmp156:
					tmp151 = Child()
					offset = tmp151.deserialize(s, offset)
				else:
					tmp151 = None
				self.v22[tmp150] = tmp151
		else:
			self.v22 = None
		
		# deserialize self.v23
		tmp157 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp157:
			self.v23 = [[None for _ in range(5)] for _ in range(10)]
			for tmp158 in range(5):
				for tmp159 in range(10):
					tmp160 = struct.unpack('B', s[offset:offset + 1])[0]
					offset += 1
					if tmp160:
						self.v23[tmp158][tmp159] = Child()
						offset = self.v23[tmp158][tmp159].deserialize(s, offset)
					else:
						self.v23[tmp158][tmp159] = None
		else:
			self.v23 = None
		
		return offset



# -*- coding: utf-8 -*-

# python imports
import sys
import struct
from enum import Enum

PY3 = sys.version_info > (3,)


class Parent1(object):

	@staticmethod
	def name():
		return 'Parent1'


	def __init__(self, count=None):
		self.initialize(count)
	

	def initialize(self, count=None):
		self.count = count
	

	def serialize(self):
		s = b''
		
		# serialize self.count
		s += b'\x00' if self.count is None else b'\x01'
		if self.count is not None:
			s += struct.pack('I', self.count)
		
		return s
	

	def deserialize(self, s, offset=0):
		# deserialize self.count
		tmp0 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp0:
			self.count = struct.unpack('I', s[offset:offset + 4])[0]
			offset += 4
		else:
			self.count = None
		
		return offset


class Parent2(object):

	@staticmethod
	def name():
		return 'Parent2'


	def __init__(self, number=None):
		self.initialize(number)
	

	def initialize(self, number=None):
		self.number = number
	

	def serialize(self):
		s = b''
		
		# serialize self.number
		s += b'\x00' if self.number is None else b'\x01'
		if self.number is not None:
			s += struct.pack('q', self.number)
		
		return s
	

	def deserialize(self, s, offset=0):
		# deserialize self.number
		tmp1 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp1:
			self.number = struct.unpack('q', s[offset:offset + 8])[0]
			offset += 8
		else:
			self.number = None
		
		return offset


class Parent3(object):

	@staticmethod
	def name():
		return 'Parent3'


	def __init__(self):
		self.initialize()
	

	def initialize(self):
		pass
	

	def serialize(self):
		s = b''
		
		return s
	

	def deserialize(self, s, offset=0):
		return offset


class Child(Parent1, Parent2):

	@staticmethod
	def name():
		return 'Child'


	def __init__(self, count=None, number=None, firstname=None):
		self.initialize(count, number, firstname)
	

	def initialize(self, count=None, number=None, firstname=None):
		Parent1.initialize(self, count)
		Parent2.initialize(self, number)
		
		self.firstname = firstname
	

	def serialize(self):
		s = b''
		
		# serialize parents
		s += Parent1.serialize(self)
		s += Parent2.serialize(self)
		
		# serialize self.firstname
		s += b'\x00' if self.firstname is None else b'\x01'
		if self.firstname is not None:
			tmp2 = b''
			tmp2 += struct.pack('I', len(self.firstname))
			while len(tmp2) and tmp2[-1] == b'\x00'[0]:
				tmp2 = tmp2[:-1]
			s += struct.pack('B', len(tmp2))
			s += tmp2
			
			s += self.firstname.encode('ISO-8859-1') if PY3 else self.firstname
		
		return s
	

	def deserialize(self, s, offset=0):
		# deserialize parents
		offset = Parent1.deserialize(self, s, offset)
		offset = Parent2.deserialize(self, s, offset)
		
		# deserialize self.firstname
		tmp3 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp3:
			tmp4 = struct.unpack('B', s[offset:offset + 1])[0]
			offset += 1
			tmp5 = s[offset:offset + tmp4]
			offset += tmp4
			tmp5 += b'\x00' * (4 - tmp4)
			tmp6 = struct.unpack('I', tmp5)[0]
			
			self.firstname = s[offset:offset + tmp6].decode('ISO-8859-1') if PY3 else s[offset:offset + tmp6]
			offset += tmp6
		else:
			self.firstname = None
		
		return offset


class GrandChild(Child, Parent3):

	@staticmethod
	def name():
		return 'GrandChild'


	def __init__(self, count=None, number=None, firstname=None, height=None):
		self.initialize(count, number, firstname, height)
	

	def initialize(self, count=None, number=None, firstname=None, height=None):
		Child.initialize(self, count, number, firstname)
		Parent3.initialize(self)
		
		self.height = height
	

	def serialize(self):
		s = b''
		
		# serialize parents
		s += Child.serialize(self)
		s += Parent3.serialize(self)
		
		# serialize self.height
		s += b'\x00' if self.height is None else b'\x01'
		if self.height is not None:
			s += struct.pack('f', self.height)
		
		return s
	

	def deserialize(self, s, offset=0):
		# deserialize parents
		offset = Child.deserialize(self, s, offset)
		offset = Parent3.deserialize(self, s, offset)
		
		# deserialize self.height
		tmp7 = struct.unpack('B', s[offset:offset + 1])[0]
		offset += 1
		if tmp7:
			self.height = struct.unpack('f', s[offset:offset + 4])[0]
			offset += 4
		else:
			self.height = None
		
		return offset



import sys
from setuptools import setup, find_packages


setup(
    name='koala-serializer',
    version='0.8.0',
    description='A tool designed for creating serializable objects among different programming languages',
    long_description='',
    author='Koala',
    author_email='mdan.hagh@gmail.com',
    url='https://github.com/koala-team/serializer',
    keywords='serialize deserialize object compile binary',

    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Topic :: Utilities'
    ],

    license='MIT',

    install_requires=[
        'configparser==3.5.0',
        'enum34==1.1.6'
    ],

    packages=find_packages(),

    # Make this executable from command line when installed
    scripts=['koalasc'],
    provides=['koalasc']
)



# -*- coding: utf-8 -*-

# python imports
import warnings

# project imports
from ..parser_core import Tokens



class RootGenerator:

    def __init__(self, code_editor):
        self._code_editor = code_editor


    def gen_root(self):
        code_editor = self._code_editor

        self._gen_imports()
        code_editor.add_line()
        self._gen_namespace('KS')
        self._gen_ks_object()
        self._code_editor.add_line("} // namespace KS")


    def _gen_imports(self):
        self._code_editor.add_line("using System;")


    def _gen_ks_object(self):
        self._code_editor.add_lines([
            "public abstract partial class KSObject",
            "{",
            "\tpublic const string NameStatic = \"\";",
            "\tpublic abstract string Name();",
            "\tpublic abstract byte[] Serialize();",
            "\tpublic abstract uint Deserialize(byte[] s, uint offset = 0);",
            "}",
        ], local_indent=1)


    def _gen_namespace(self, ns_name):
        self._code_editor.add_line("namespace %s" % ns_name)
        self._code_editor.add_line('{')


###################################################################################
###################################################################################


class HeaderGenerator:

    def __init__(self, code_editor):
        self._code_editor = code_editor


    def gen_header(self, module_name):
        code_editor = self._code_editor

        self._gen_imports()
        code_editor.add_line()
        self._gen_namespace("KS." + module_name)
        code_editor.increase_indentation(put_accolade=False)


    def _gen_imports(self):
        self._code_editor.add_line("using System;")
        self._code_editor.add_line("using System.Linq;")
        self._code_editor.add_line("using System.Collections.Generic;")


    def _gen_namespace(self, ns_name):
        self._code_editor.add_line("namespace %s" % ns_name)
        self._code_editor.add_line('{')


###################################################################################
###################################################################################


class FooterGenerator:

    def __init__(self, code_editor):
        self._code_editor = code_editor


    def gen_footer(self, module_name):
        self._code_editor.decrease_indentation(put_accolade=False)
        self._gen_namespace("KS." + module_name)


    def _gen_namespace(self, ns_name):
        self._code_editor.add_line("} // namespace %s" % ns_name)


###################################################################################
###################################################################################


class TypeGenerator:

    def __init__(self, code_editor):
        self._user_defined_types = {
            Tokens.Class: {},
            Tokens.Enum: {}
        }

        self._code_editor = code_editor
        self._serializer_generator = SerializerGenerator(self._user_defined_types, code_editor)
        self._deserializer_generator = DeserializerGenerator(self._user_defined_types, code_editor)
        self._variable_type_generator = VariableTypeGenerator(self._user_defined_types)


    def gen_type(self, type_name, properties):
        gen_func = getattr(self, '_gen_type_%s' % properties['_def'][0])
        gen_func(type_name, properties)



    def _gen_type_class(self, type_name, properties):
        code_editor = self._code_editor

        _def = properties.pop('_def')
        properties = properties.items()

        parents = ['KSObject'] if len(_def) == 1 else [p[0] for p in _def[1]]
        if len(parents) > 1:
            warnings.warn("C# does not support multiple inheritance. Only first parent will be considered.")
            parents = parents[:1]
        self._user_defined_types[Tokens.Class][type_name] = (parents, properties)

        # generate definitions
        code_editor.add_line("public partial class %s : %s" % (type_name, ', '.join(parents)))
        code_editor.increase_indentation()
        for attr_name, tree in properties:
            vartype = self._variable_type_generator.gen_vartype(tree)
            code_editor.add_line("public %s %s { get; set; }" % (vartype, attr_name))

        code_editor.add_line('\n')

        # generate constructor
        code_editor.add_line("public %s()" % type_name)
        code_editor.increase_indentation()
        code_editor.decrease_indentation()
        code_editor.add_line()

        # generate name
        code_editor.add_line("public new const string NameStatic = \"%s\";" % type_name)
        code_editor.add_line()

        code_editor.add_line("public override string Name() => \"%s\";" % type_name)
        code_editor.add_line()

        # generate serializer
        result_name = 's'
        code_editor.add_line("public override byte[] Serialize()")
        code_editor.increase_indentation()
        code_editor.add_line("List<byte> %s = new List<byte>();" % result_name)
        code_editor.add_line()

        if parents[0] != 'KSObject':
            code_editor.add_line("// serialize parents")
            # for parent in parents:
            #     code_editor.add_line("%s.AddRange(%s.Serialize());" % (result_name, parent))
            code_editor.add_line("%s.AddRange(base.Serialize());" % result_name)
            code_editor.add_line()

        for attr_name, tree in properties:
            code_editor.add_line("// serialize %s" % attr_name)
            self._serializer_generator.gen_serializer(attr_name, result_name, tree)
            code_editor.add_line()

        code_editor.add_line("return %s.ToArray();" % result_name)
        code_editor.decrease_indentation()
        code_editor.add_line()

        # generate deserializer
        data_name = 's'
        offset_name = 'offset'
        code_editor.add_line("public override uint Deserialize(byte[] %s, uint %s = 0)" %
                             (data_name, offset_name))
        code_editor.increase_indentation()

        if parents[0] != 'KSObject':
            code_editor.add_line("// deserialize parents")
            # for parent in parents:
            #     code_editor.add_line("%s = %s.Deserialize(%s, %s);" % 
            #                          (offset_name, parent, data_name, offset_name))
            code_editor.add_line("%s = base.Deserialize(%s, %s);" % 
                                 (offset_name, data_name, offset_name))
            code_editor.add_line()

        for attr_name, tree in properties:
            code_editor.add_line("// deserialize %s" % attr_name)
            self._deserializer_generator.gen_deserializer(data_name, offset_name, attr_name, tree)
            code_editor.add_line()

        code_editor.add_line("return %s;" % offset_name)
        code_editor.decrease_indentation()

        code_editor.decrease_indentation()


    def _gen_type_enum(self, type_name, properties):
        code_editor = self._code_editor

        _def = properties.pop('_def')
        value_type = _def[1][0]
        self._user_defined_types[Tokens.Enum][type_name] = value_type
        code_editor.add_line("public enum %s" % type_name)
        code_editor.increase_indentation()

        value = 0
        for attr in _def[1][1]:
            name = attr[0]
            if len(attr) > 1:
                value = attr[1]
            code_editor.add_line("%s = %s," % (name, value))
            value += 1

        code_editor.decrease_indentation()

###################################################################################
###################################################################################


class SerializerGenerator:

    def __init__(self, user_defined_types, code_editor):
        self._simples_token_table = {
            Tokens.Boolean: 'bool',
            Tokens.Char: 'char',
            Tokens.Byte: 'sbyte',
            Tokens.UnsignedByte: 'byte',
            Tokens.Short: 'short',
            Tokens.UnsignedShort: 'ushort',
            Tokens.Integer: 'int',
            Tokens.UnsignedInteger: 'uint',
            Tokens.Long: 'long',
            Tokens.UnsignedLong: 'ulong',
            Tokens.Float: 'float',
            Tokens.Double: 'double'
        }

        self._user_defined_types = user_defined_types
        self._code_editor = code_editor


    def gen_serializer(self, value_name, result_name, tree):
        value_type = tree[0]
        serializer_func = None

        for type_token in self._user_defined_types:
            if value_type in self._user_defined_types[type_token]:
                serializer_func = getattr(self, '_gen_serializer_%s' % type_token)
                break

        if serializer_func is None:
            serializer_func = getattr(
                self,
                '_gen_serializer_%s' % value_type,
                self._gen_serializer_simples
            )

        self._code_editor.add_line("%s.Add((byte)((%s == null) ? 0 : 1));" % 
                                   (result_name, value_name))
        self._code_editor.add_line("if (%s != null)" % value_name)
        self._code_editor.increase_indentation()
        serializer_func(value_name, result_name, tree)
        self._code_editor.decrease_indentation()



    def _gen_serializer_class(self, value_name, result_name, tree):
        self._code_editor.add_line("%s.AddRange(%s.Serialize());" % (result_name, value_name))


    def _gen_serializer_enum(self, value_name, result_name, tree):
        value_type = self._user_defined_types[Tokens.Enum][tree[0]][0]
        value_type = self._simples_token_table[value_type]
        self._gen_serializer_simples(
            "((%s)%s)" % (value_type, value_name),
            result_name,
            self._user_defined_types[Tokens.Enum][tree[0]]
        )


    def _gen_serializer_list(self, value_name, result_name, tree):
        code_editor = self._code_editor

        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen items
        item = code_editor.new_tempvar()
        code_editor.add_line("foreach (var %s in %s)" % (item, value_name))
        code_editor.increase_indentation()
        self.gen_serializer(item, result_name, tree[1])
        code_editor.decrease_indentation()


    def _gen_serializer_map(self, value_name, result_name, tree):
        code_editor = self._code_editor

        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen items
        pair = code_editor.new_tempvar()
        code_editor.add_line("foreach (var %s in %s)" % (pair, value_name))
        code_editor.increase_indentation()
        self.gen_serializer("%s.Key" % pair, result_name, tree[1][0])
        code_editor.add_line()
        self.gen_serializer("%s.Value" % pair, result_name, tree[1][1])
        code_editor.decrease_indentation()


    def _gen_serializer_array(self, value_name, result_name, tree):
        code_editor = self._code_editor

        dims = tree[1][0]
        array_type = tree[1][1]

        indexes = []
        for dim in dims:
            index = code_editor.new_tempvar()
            indexes.append(index)
            code_editor.add_line("for (uint %s = 0; %s < %s; %s++)" %
                                 (index, index, dim, index))
            code_editor.increase_indentation()

        value_name += "[%s]" % ', '.join(indexes)
        self.gen_serializer(value_name, result_name, array_type)
        for _ in range(len(dims)):
            code_editor.decrease_indentation()


    def _gen_serializer_string(self, value_name, result_name, tree):
        # gen size
        self._gen_size(value_name, result_name, tree)

        # gen string
        self._code_editor.add_line("%s.AddRange(System.Text.Encoding.GetEncoding(\"ISO-8859-1\").GetBytes(%s));" %
                                    (result_name, value_name))


    def _gen_serializer_simples(self, value_name, result_name, tree):
        if tree[0] == Tokens.UnsignedByte:
            self._code_editor.add_line("%s.Add((byte)%s);" % (result_name, value_name))
        elif tree[0] == Tokens.Byte:
            self._code_editor.add_line("%s.Add((byte)%s);" % (result_name, value_name))
        else:
            value_type = self._simples_token_table[tree[0]]
            self._code_editor.add_line("%s.AddRange(BitConverter.GetBytes((%s)%s));" % 
                                        (result_name, value_type, value_name))


    def _gen_size(self, value_name, result_name, tree):
        code_editor = self._code_editor

        size = code_editor.new_tempvar()
        code_editor.add_line("List<byte> %s = new List<byte>();" % size)
        self._gen_serializer_simples("%s.Count()" % value_name, size, (Tokens.UnsignedInteger, ))
        code_editor.add_line("while (%s.Count > 0 && %s.Last() == 0)" % (size, size))
        code_editor.add_line("%s.RemoveAt(%s.Count - 1);" % (size, size), 1)

        self._gen_serializer_simples("%s.Count" % size, result_name, (Tokens.UnsignedByte, ))
        code_editor.add_line("%s.AddRange(%s);" % (result_name, size))
        code_editor.add_line()


###################################################################################
###################################################################################


class DeserializerGenerator:

    def __init__(self, user_defined_types, code_editor):
        self._simples_method_table = {
            Tokens.Boolean: 'ToBoolean',
            Tokens.Char: 'ToChar',
            Tokens.Byte: None,
            Tokens.UnsignedByte: None,
            Tokens.Short: 'ToInt16',
            Tokens.UnsignedShort: 'ToUInt16',
            Tokens.Integer: 'ToInt32',
            Tokens.UnsignedInteger: 'ToUInt32',
            Tokens.Long: 'ToInt64',
            Tokens.UnsignedLong: 'ToUInt64',
            Tokens.Float: 'ToSingle',
            Tokens.Double: 'ToDouble'
        }
        self._simples_token_table = {
            Tokens.Boolean: 'bool',
            Tokens.Char: 'char',
            Tokens.Byte: 'sbyte',
            Tokens.UnsignedByte: 'byte',
            Tokens.Short: 'short',
            Tokens.UnsignedShort: 'ushort',
            Tokens.Integer: 'int',
            Tokens.UnsignedInteger: 'uint',
            Tokens.Long: 'long',
            Tokens.UnsignedLong: 'ulong',
            Tokens.Float: 'float',
            Tokens.Double: 'double'
        }

        self._user_defined_types = user_defined_types
        self._code_editor = code_editor
        self._variable_type_generator = VariableTypeGenerator(self._user_defined_types)


    def gen_deserializer(self, data_name, offset_name, value_name, tree):
        value_type = tree[0]
        deserializer_func = None

        for type_token in self._user_defined_types:
            if value_type in self._user_defined_types[type_token]:
                deserializer_func = getattr(self, '_gen_deserializer_%s' % type_token)
                break

        if deserializer_func is None:
            deserializer_func = getattr(
                self,
                '_gen_deserializer_%s' % value_type,
                self._gen_deserializer_simples
            )

        
        is_null = self._code_editor.new_tempvar()
        self._code_editor.add_line("byte %s;" % is_null)
        self._gen_deserializer_simples(data_name, offset_name, is_null, (Tokens.UnsignedByte, ))
        self._code_editor.add_line("if (%s == 1)" % is_null)
        self._code_editor.increase_indentation()
        deserializer_func(data_name, offset_name, value_name, tree)
        self._code_editor.decrease_indentation()
        self._code_editor.add_line("else")
        self._code_editor.add_line("%s = null;" % value_name, 1)



    def _gen_deserializer_class(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        code_editor.add_line("%s = new %s();" % (value_name, tree[0]))
        code_editor.add_line("%s = %s.Deserialize(%s, %s);" % 
                             (offset_name, value_name, data_name, offset_name))


    def _gen_deserializer_enum(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        enum_type = tree[0]
        val = code_editor.new_tempvar()
        val_type = self._user_defined_types[Tokens.Enum][tree[0]][0]
        val_type = self._simples_token_table[val_type]
        code_editor.add_line("%s %s;" % (val_type, val))

        self._gen_deserializer_simples(
            data_name,
            offset_name,
            val,
            self._user_defined_types[Tokens.Enum][tree[0]]
        )
        code_editor.add_line("%s = (%s)%s;" % (value_name, enum_type, val))


    def _gen_deserializer_list(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen items
        value_type = self._variable_type_generator.gen_vartype(tree)
        code_editor.add_line("%s = new %s();" % (value_name, value_type))
        i = code_editor.new_tempvar()
        code_editor.add_line("for (uint %s = 0; %s < %s; %s++)" % (i, i, size, i))
        code_editor.increase_indentation()
        item = code_editor.new_tempvar()
        item_type = self._variable_type_generator.gen_vartype(tree[1])
        code_editor.add_line("%s %s;" % (item_type, item))
        self.gen_deserializer(data_name, offset_name, item, tree[1])
        code_editor.add_line("%s.Add(%s);" % (value_name, item))
        code_editor.decrease_indentation()


    def _gen_deserializer_map(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen items
        value_type = self._variable_type_generator.gen_vartype(tree)
        code_editor.add_line("%s = new %s();" % (value_name, value_type))
        i = code_editor.new_tempvar()
        code_editor.add_line("for (uint %s = 0; %s < %s; %s++)" % (i, i, size, i))
        code_editor.increase_indentation()

        key = code_editor.new_tempvar()
        val = code_editor.new_tempvar()
        key_type = self._variable_type_generator.gen_vartype(tree[1][0])
        val_type = self._variable_type_generator.gen_vartype(tree[1][1])

        code_editor.add_line("%s %s;" % (key_type, key))
        self.gen_deserializer(data_name, offset_name, key, tree[1][0])
        code_editor.add_line()

        code_editor.add_line("%s %s;" % (val_type, val))
        self.gen_deserializer(data_name, offset_name, val, tree[1][1])
        code_editor.add_line()

        code_editor.add_line("%s[%s] = %s;" % (value_name, key, val))
        code_editor.decrease_indentation()


    def _gen_deserializer_array(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        dims = tree[1][0]
        array_type_tree = tree[1][1]
        array_type = self._variable_type_generator.gen_vartype(array_type_tree)
        code_editor.add_line("%s = new %s[%s];" %
                             (value_name, array_type, ', '.join([str(dim) for dim in dims])))

        indexes = []
        for dim in dims:
            index = code_editor.new_tempvar()
            indexes.append(index)
            code_editor.add_line("for (uint %s = 0; %s < %s; %s++)" %
                                 (index, index, dim, index))
            code_editor.increase_indentation()

        value_name += "[%s]" % ', '.join(indexes)
        self.gen_deserializer(data_name, offset_name, value_name, array_type_tree)
        for _ in range(len(dims)):
            code_editor.decrease_indentation()


    def _gen_deserializer_string(self, data_name, offset_name, value_name, tree):
        # gen size
        size = self._gen_size(data_name, offset_name, value_name, tree)

        # gen string
        self._code_editor.add_line("%s = System.Text.Encoding.GetEncoding(\"ISO-8859-1\").GetString(%s.Skip((int)%s).Take((int)%s).ToArray());" %
                                   (value_name, data_name, offset_name, size))
        self._code_editor.add_line("%s += %s;" % (offset_name, size))


    def _gen_deserializer_simples(self, data_name, offset_name, value_name, tree):
        method = self._simples_method_table[tree[0]]
        value_type = self._simples_token_table[tree[0]]
        if tree[0] in [Tokens.UnsignedByte, Tokens.Byte]:
            self._code_editor.add_line("%s = (%s)%s[(int)%s];" %
                                       (value_name, value_type, data_name, offset_name or 0))
        else:
            self._code_editor.add_line("%s = BitConverter.%s(%s, (int)%s);" %
                                        (value_name, method, data_name, offset_name or 0))
        if offset_name is not None:
            self._code_editor.add_line("%s += sizeof(%s);" % (offset_name, value_type))


    def _gen_size(self, data_name, offset_name, value_name, tree):
        code_editor = self._code_editor

        size_len = code_editor.new_tempvar()
        code_editor.add_line("byte %s;" % size_len)
        self._gen_deserializer_simples(data_name, offset_name, size_len, (Tokens.UnsignedByte, ))
        size_bytes = code_editor.new_tempvar()
        code_editor.add_line("byte[] %s = new byte[sizeof(%s)];" %
                             (size_bytes, Tokens.UnsignedInteger))
        code_editor.add_line("Array.Copy(%s, %s, %s, 0, %s);" %
                             (data_name, offset_name, size_bytes, size_len))
        code_editor.add_line("%s += %s;" % (offset_name, size_len))

        size = code_editor.new_tempvar()
        code_editor.add_line("uint %s;" % size)
        self._gen_deserializer_simples(size_bytes, None, size, (Tokens.UnsignedInteger, ))
        code_editor.add_line()

        return size


###################################################################################
###################################################################################


class VariableTypeGenerator:

    def __init__(self, user_defined_types):
        self._user_defined_types = user_defined_types

        self._simples_token_table = {
            Tokens.Boolean: 'bool?',
            Tokens.Char: 'char?',
            Tokens.Byte: 'sbyte?',
            Tokens.UnsignedByte: 'byte?',
            Tokens.Short: 'short?',
            Tokens.UnsignedShort: 'ushort?',
            Tokens.Integer: 'int?',
            Tokens.UnsignedInteger: 'uint?',
            Tokens.Long: 'long?',
            Tokens.UnsignedLong: 'ulong?',
            Tokens.Float: 'float?',
            Tokens.Double: 'double?',
            Tokens.String: 'string'
        }


    def gen_vartype(self, tree):
        vartype = tree[0]

        for type_token in self._user_defined_types:
            if vartype in self._user_defined_types[type_token]:
                nullable = '?' if type_token == Tokens.Enum else ''
                return vartype + nullable

        gen_vartype_func = getattr(
            self,
            '_gen_vartype_%s' % vartype,
            self._gen_vartype_simples
        )

        return gen_vartype_func(tree)


    def _gen_vartype_list(self, tree):
        return "List<%s>" % self.gen_vartype(tree[1])


    def _gen_vartype_map(self, tree):
        first_type = self.gen_vartype(tree[1][0])
        second_type = self.gen_vartype(tree[1][1])
        return "Dictionary<%s, %s>" % (first_type, second_type)


    def _gen_vartype_array(self, tree):
        dims = list(reversed(tree[1][0]))
        initial_type = self.gen_vartype(tree[1][1])

        array_type = initial_type
        dim = ',' * (len(dims) - 1)
        array_type = "%s[%s]" % (array_type, dim)
        return array_type


    def _gen_vartype_simples(self, tree):
        return self._simples_token_table[tree[0]]



# -*- coding: utf-8 -*-

# python imports
import configparser
from collections import OrderedDict

# project imports
from .rule_parser import RuleParser


class Parser:

    def __init__(self, source_path):
        self._src = configparser.ConfigParser()
        self._src.optionxform = str
        with open(source_path) as f:
            self._src.readfp(f)

        self._rule_parser = RuleParser()
        self._parse_tree = self._parse()


    def parse_tree(self):
        return self._parse_tree


    def _parse(self):
        src = self._src
        parse_tree = []

        user_defined_types = src.sections()
        for udt in user_defined_types:
            attr_trees = OrderedDict()
            for attr, type in src.items(udt):
                attr_trees[attr] = self._rule_parser.get_type_tree(type, udt, user_defined_types)
            parse_tree.append((udt, attr_trees))

        return parse_tree
